import logging
from abc import ABC
from enum import Enum
from typing import Dict, List, Literal, Optional

from pydantic import Field
from pydantic.class_validators import root_validator, validator
from pydantic.types import NonNegativeFloat, PositiveInt

from hydrolib.core.basemodel import DiskOnlyFileModel
from hydrolib.core.dflowfm.common import LocationType
from hydrolib.core.dflowfm.common.models import Operand
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    get_enum_validator,
    make_list_validator,
    validate_required_fields,
)

logger = logging.getLogger(__name__)


class DataFileType(str, Enum):
    """
    Enum class containing the valid values for the dataFileType
    attribute in several subclasses of AbstractIniField.
    """

    arcinfo = "arcinfo"
    geotiff = "GeoTIFF"
    sample = "sample"
    onedfield = "1dField"
    polygon = "polygon"

    allowedvaluestext = "Possible values: arcinfo, GeoTIFF, sample, 1dField, polygon."


class InterpolationMethod(str, Enum):
    """
    Enum class containing the valid values for the interpolationMethod
    attribute in several subclasses of AbstractIniField.
    """

    constant = "constant"  # only with dataFileType=polygon .
    triangulation = "triangulation"  # Delaunay triangulation+linear interpolation.
    averaging = "averaging"  # grid cell averaging.

    allowedvaluestext = "Possible values: constant, triangulation, averaging."


class AveragingType(str, Enum):
    """
    Enum class containing the valid values for the averagingType
    attribute in several subclasses of AbstractIniField.
    """

    mean = "mean"  # simple average
    nearestnb = "nearestNb"  # nearest neighbour value
    max = "max"  # highest
    min = "min"  # lowest
    invdist = "invDist"  # inverse-weighted distance average
    minabs = "minAbs"  # smallest absolute value

    allowedvaluestext = "Possible values: mean, nearestNb, max, min, invDist, minAbs."


class IniFieldGeneral(INIGeneral):
    """The initial field file's `[General]` section with file meta data."""

    class Comments(INIBasedModel.Comments):
        fileversion: Optional[str] = Field(
            "File version. Do not edit this.", alias="fileVersion"
        )
        filetype: Optional[str] = Field(
            "File type. Should be 'iniField'. Do not edit this.",
            alias="fileType",
        )

    comments: Comments = Comments()
    _header: Literal["General"] = "General"
    fileversion: str = Field("2.00", alias="fileVersion")
    filetype: Literal["iniField"] = Field("iniField", alias="fileType")


class AbstractSpatialField(INIBasedModel, ABC):
    """
    Abstract base class for `[Initial]` and `[Parameter]` block data in
    inifield files.

    Defines all common fields. Used via subclasses InitialField and ParameterField.
    """

    class Comments(INIBasedModel.Comments):
        quantity: Optional[str] = Field(
            "Name of the quantity. See UM Table D.2.", alias="quantity"
        )
        datafile: Optional[str] = Field(
            "Name of file containing field data values.", alias="dataFile"
        )
        datafiletype: Optional[str] = Field("Type of dataFile.", alias="dataFileType")
        interpolationmethod: Optional[str] = Field(
            "Type of (spatial) interpolation.", alias="interpolationmethod"
        )
        operand: Optional[str] = Field(
            "How this data is combined with previous data for the same quantity (if any).",
            alias="operand",
        )
        averagingtype: Optional[str] = Field(
            "Type of averaging, if interpolationMethod=averaging .",
            alias="averagingtype",
        )
        averagingrelsize: Optional[str] = Field(
            "Relative search cell size for averaging.", alias="averagingrelsize"
        )
        averagingnummin: Optional[str] = Field(
            "Minimum number of points in averaging. Must be ≥ 1.",
            alias="averagingnummin",
        )
        averagingpercentile: Optional[str] = Field(
            "Percentile value for which data values to include in averaging. 0.0 means off.",
            alias="averagingpercentile",
        )
        extrapolationmethod: Optional[str] = Field(
            "Option for (spatial) extrapolation.", alias="extrapolationmethod"
        )
        locationtype: Optional[str] = Field(
            "Target location of interpolation.", alias="locationtype"
        )
        value: Optional[str] = Field(
            "Only for dataFileType=polygon. The constant value to be set inside for all model points inside the polygon."
        )

    comments: Comments = Comments()

    quantity: str = Field(alias="quantity")
    datafile: DiskOnlyFileModel = Field(alias="dataFile")

    datafiletype: DataFileType = Field(alias="dataFileType")
    interpolationmethod: Optional[InterpolationMethod] = Field(
        alias="interpolationMethod"
    )
    operand: Optional[Operand] = Field(Operand.override.value, alias="operand")
    averagingtype: Optional[AveragingType] = Field(
        AveragingType.mean.value, alias="averagingType"
    )
    averagingrelsize: Optional[NonNegativeFloat] = Field(1.01, alias="averagingRelSize")
    averagingnummin: Optional[PositiveInt] = Field(1, alias="averagingNumMin")
    averagingpercentile: Optional[NonNegativeFloat] = Field(
        0, alias="averagingPercentile"
    )
    extrapolationmethod: Optional[bool] = Field(False, alias="extrapolationMethod")
    locationtype: Optional[LocationType] = Field(
        LocationType.all.value, alias="locationType"
    )
    value: Optional[float] = Field(alias="value")

    datafiletype_validator = get_enum_validator("datafiletype", enum=DataFileType)
    interpolationmethod_validator = get_enum_validator(
        "interpolationmethod", enum=InterpolationMethod
    )
    operand_validator = get_enum_validator("operand", enum=Operand)
    averagingtype_validator = get_enum_validator("averagingtype", enum=AveragingType)
    locationtype_validator = get_enum_validator("locationtype", enum=LocationType)

    @root_validator(allow_reuse=True)
    def validate_that_value_is_present_for_polygons(cls, values: Dict) -> Dict:
        """Validates that the value is provided when dealing with polygons."""
        return validate_required_fields(
            values,
            "value",
            conditional_field_name="datafiletype",
            conditional_value=DataFileType.polygon,
        )

    @validator("value", always=True)
    @classmethod
    def _validate_value_and_filetype(cls, v, values: dict):
        if v is not None and values.get("datafiletype") != DataFileType.polygon:
            raise ValueError(
                f"When value={v} is given, dataFileType={DataFileType.polygon} is required."
            )

        return v


class InitialField(AbstractSpatialField):
    """
    Initial condition field definition, represents an `[Initial]` block in
    an inifield file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile.initial[..]`
    """

    _header: Literal["Initial"] = "Initial"


class ParameterField(AbstractSpatialField):
    """
    Parameter field definition, represents a `[Parameter]` block in
    an inifield file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile.parameter[..]`
    """

    _header: Literal["Parameter"] = "Parameter"


class IniFieldModel(INIModel):
    """
    The overall inifield model that contains the contents of one initial field and parameter file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile[..]`.

    Attributes:
        general (IniFieldGeneral): `[General]` block with file metadata.
        initial (List[InitialField]): List of `[Initial]` blocks with initial condition definitions.
        parameter (List[ParameterField]): List of `[Parameter]` blocks with spatial parameter definitions.
    """

    general: IniFieldGeneral = IniFieldGeneral()
    initial: List[InitialField] = []
    parameter: List[ParameterField] = []

    _split_to_list = make_list_validator("initial", "parameter")

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "fieldFile"



from __future__ import annotations

import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

import meshkernel as mk
import netCDF4 as nc
import numpy as np
import numpy.typing as npt
from meshkernel.py_structures import GeometryList
from pydantic import Field

from hydrolib.core import __version__
from hydrolib.core.basemodel import (
    BaseModel,
    ModelSaveSettings,
    ParsableFileModel,
    file_load_context,
)
from hydrolib.core.dflowfm.net.reader import UgridReader
from hydrolib.core.dflowfm.net.writer import UgridWriter

logger = logging.getLogger(__name__)


def split_by(gl: mk.GeometryList, by: float) -> list:
    """Function to split mk.GeometryList by seperator.

    Args:
        gl (mk.GeometryList): The geometry list to split.
        by (float): The value by which to split the gl.

    Returns:
        list: The split lists.
    """
    x, y = gl.x_coordinates.copy(), gl.y_coordinates.copy()
    idx = np.where(x == by)[0]

    xparts = np.split(x, idx)
    yparts = np.split(y, idx)

    lists = [
        mk.GeometryList(xp[min(i, 1) :], yp[min(i, 1) :])
        for i, (xp, yp) in enumerate(zip(xparts, yparts))
    ]

    return lists


class Mesh2d(BaseModel):
    """Mesh2d defines a single two dimensional grid.

    Attributes:
        meshkernel (mk.MeshKernel):
            The meshkernel used to manimpulate this Mesh2d.
        mesh2d_node_z (np.ndarray):
            The node positions on the z-axis. Defaults to np.empty(0, dtype=np.double).
        mesh2d_face_z (np.ndarray):
            The face positions on the z-axis. Defaults to np.empty(0, dtype=np.double).
    """

    meshkernel: mk.MeshKernel = Field(default_factory=mk.MeshKernel)

    # placeholders for bathymetry
    mesh2d_node_z: np.ndarray = Field(
        default_factory=lambda: np.empty(0, dtype=np.double)
    )
    mesh2d_face_z: np.ndarray = Field(
        default_factory=lambda: np.empty(0, dtype=np.double)
    )

    @property
    def mesh2d_node_x(self) -> np.ndarray[float]:
        """The x-coordinates of the nodes in the mesh.

        Returns:
            ndarray[float]: A 1D double array describing the x-coordinates of the nodes.
        """
        return self.meshkernel.mesh2d_get().node_x

    @property
    def mesh2d_node_y(self) -> np.ndarray[float]:
        """The y-coordinates of the nodes in the mesh.

        Returns:
            ndarray[float]: A 1D double array describing the y-coordinates of the nodes.
        """
        return self.meshkernel.mesh2d_get().node_y

    @property
    def mesh2d_edge_x(self) -> np.ndarray[float]:
        """The x-coordinates of the mesh edges' middle points.

        Returns:
            ndarray[float]: A 1D double array describing x-coordinates of the mesh edges' middle points.
        """
        return self.meshkernel.mesh2d_get().edge_x

    @property
    def mesh2d_edge_y(self) -> np.ndarray[float]:
        """The y-coordinates of the mesh edges' middle points.

        Returns:
            ndarray[float]: A 1D double array describing y-coordinates of the mesh edges' middle points.
        """
        return self.meshkernel.mesh2d_get().edge_y

    @property
    def mesh2d_edge_nodes(self) -> np.ndarray[int, int]:
        """The node indices of the mesh edges.

        Returns:
            np.ndarray[int, int]: A 2D integer array (nEdges, 2) containg the two node indices for each edge.
        """
        mesh2d_output = self.meshkernel.mesh2d_get()
        edge_nodes = mesh2d_output.edge_nodes.reshape((-1, 2))
        return edge_nodes

    @property
    def mesh2d_face_x(self) -> np.ndarray[float]:
        """The x-coordinates of the mesh faces' mass centers.

        Returns:
            ndarray[float]: A 1D double array describing x-coordinates of the mesh faces' mass centers.
        """
        return self.meshkernel.mesh2d_get().face_x

    @property
    def mesh2d_face_y(self) -> np.ndarray[float]:
        """The y-coordinates of the mesh faces' mass centers.

        Returns:
            ndarray[float]: A 1D double array describing y-coordinates of the mesh faces' mass centers.
        """
        return self.meshkernel.mesh2d_get().face_y

    @property
    def mesh2d_face_nodes(self) -> np.ndarray[int, int]:
        """The node indices of the mesh faces

        Returns:
            np.ndarray[int, int]: A 2D integer array describing the nodes composing each mesh 2d face. A 2D integer array (nFaces, maxNodesPerFace) containg the node indices for each face.
        """
        mesh2d_output = self.meshkernel.mesh2d_get()
        npf = mesh2d_output.nodes_per_face
        if self.is_empty():
            return np.empty((0, 0), dtype=np.int32)
        face_node_connectivity = np.full(
            (len(mesh2d_output.face_x), max(npf)), np.iinfo(np.int32).min
        )
        idx = (
            np.ones_like(face_node_connectivity) * np.arange(max(npf))[None, :]
        ) < npf[:, None]
        face_node_connectivity[idx] = mesh2d_output.face_nodes
        return face_node_connectivity

    def is_empty(self) -> bool:
        """Determine whether this Mesh2d is empty.

        Returns:
            (bool): Whether this Mesh2d is empty.
        """
        return self.mesh2d_node_x.size == 0

    def read_file(self, file_path: Path) -> None:
        """Read the Mesh2d from the file at file_path.

        Args:
            file_path (Path): Path to the file to be read.
        """
        reader = UgridReader(file_path)
        reader.read_mesh2d(self)

    def _set_mesh2d(self, node_x, node_y, edge_nodes) -> None:
        mesh2d = mk.Mesh2d(
            node_x=node_x.astype(np.float64),
            node_y=node_y.astype(np.float64),
            edge_nodes=edge_nodes.ravel().astype(np.int32),
        )

        self.meshkernel.mesh2d_set(mesh2d)

    def get_mesh2d(self) -> mk.Mesh2d:
        """Get the mesh2d as represented in the MeshKernel

        Returns:
            (mk.Mesh2d): The mesh2d as represented in the MeshKernel
        """
        return self.meshkernel.mesh2d_get()

    def create_rectilinear(self, extent: tuple, dx: float, dy: float) -> None:
        """Create a rectilinear mesh within a polygon. A rectangular grid is generated within the polygon bounds

        Args:
            extent (tuple): Bounding box of mesh (left, bottom, right, top)
            dx (float): Horizontal distance
            dy (float): Vertical distance

        Raises:
            NotImplementedError: MultiPolygons
        """

        xmin, ymin, xmax, ymax = extent

        rows = int((ymax - ymin) / dy)
        columns = int((xmax - xmin) / dx)

        params = mk.MakeGridParameters(
            num_columns=columns,
            num_rows=rows,
            origin_x=xmin,
            origin_y=ymin,
            block_size_x=dx,
            block_size_y=dy,
        )

        mesh2d_input = self.meshkernel  # mk.MeshKernel()
        mesh2d_input.curvilinear_compute_rectangular_grid(params)
        mesh2d_input.curvilinear_convert_to_mesh2d()  # convert to ugrid/mesh2d

    def create_triangular(self, geometry_list: mk.GeometryList) -> None:
        """Create triangular grid within GeometryList object

        Args:
            geometry_list (mk.GeometryList): GeometryList represeting a polygon within which the mesh is generated.
        """
        # Call meshkernel
        self.meshkernel.mesh2d_make_triangular_mesh_from_polygon(geometry_list)

    def clip(
        self,
        geometrylist: mk.GeometryList,
        deletemeshoption: mk.DeleteMeshOption = mk.DeleteMeshOption.INSIDE_NOT_INTERSECTED,
        inside=False,
    ) -> None:
        """Clip the 2D mesh by a polygon. Both outside the exterior and inside the interiors is clipped

        Args:
            geometrylist (GeometryList): Polygon stored as GeometryList
            deletemeshoption (int, optional): [description]. Defaults to 1.
        """

        # For clipping outside
        if not inside:
            # Check if a multipolygon was provided when clipping outside
            if geometrylist.geometry_separator in geometrylist.x_coordinates:
                raise NotImplementedError(
                    "Deleting outside more than a single exterior (MultiPolygon) is not implemented."
                )

            # Get exterior and interiors
            parts = split_by(geometrylist, geometrylist.inner_outer_separator)

            exteriors = [parts[0]]
            interiors = parts[1:]

        # Inside
        else:
            # Check if any polygon contains holes, when clipping inside
            if geometrylist.inner_outer_separator in geometrylist.x_coordinates:
                raise NotImplementedError(
                    "Deleting inside a (Multi)Polygon with holes is not implemented."
                )

            # Get exterior and interiors
            parts = split_by(geometrylist, geometrylist.geometry_separator)

            exteriors = parts[:]
            interiors = []

        # Check if parts are closed
        for part in exteriors + interiors:
            if (part.x_coordinates[0], part.y_coordinates[0]) != (
                part.x_coordinates[-1],
                part.y_coordinates[-1],
            ):
                raise ValueError(
                    "First and last coordinate of each GeometryList part should match."
                )

        # Delete everything outside the (Multi)Polygon
        for exterior in exteriors:
            self.meshkernel.mesh2d_delete(
                geometry_list=exterior,
                delete_option=deletemeshoption,
                invert_deletion=not inside,
            )

        # Delete all holes.
        for interior in interiors:
            self.meshkernel.mesh2d_delete(
                geometry_list=interior,
                delete_option=deletemeshoption,
                invert_deletion=inside,
            )

    def refine(self, polygon: mk.GeometryList, level: int, min_edge_size=10.0):
        """Refine the mesh within a polygon, by a number of steps (level)

        Args:
            polygon (GeometryList): Polygon in which to refine
            level (int): Number of refinement steps
        """

        # Check if parts are closed
        # if not (polygon.x_coordinates[0], polygon.y_coordinates[0]) == (
        #     polygon.x_coordinates[-1],
        #     polygon.y_coordinates[-1],
        # ):
        #     raise ValueError("First and last coordinate of each GeometryList part should match.")

        parameters = mk.MeshRefinementParameters(
            refine_intersected=True,
            use_mass_center_when_refining=False,
            min_edge_size=min_edge_size,  # Does nothing?
            refinement_type=1,  # No effect?
            connect_hanging_nodes=True,
            account_for_samples_outside_face=False,
            max_refinement_iterations=level,
        )
        self.meshkernel.mesh2d_refine_based_on_polygon(polygon, parameters)


class Branch:
    def __init__(
        self,
        geometry: np.ndarray,
        branch_offsets: np.ndarray = None,
        mask: np.ndarray = None,
    ) -> None:
        # Check that the array has two collumns (x and y)
        assert geometry.shape[1] == 2

        # Split in x and y
        self.geometry = geometry
        self._x_coordinates = geometry[:, 0]
        self._y_coordinates = geometry[:, 1]

        # Calculate distance of coordinates along line
        segment_distances = np.hypot(
            np.diff(self._x_coordinates), np.diff(self._y_coordinates)
        )
        self._distance = np.concatenate([[0], np.cumsum(segment_distances)])
        self.length = segment_distances.sum()

        # Check if mask and branch offsets (if both given) have same shape
        if (
            mask is not None
            and branch_offsets is not None
            and branch_offsets.shape != mask.shape
        ):
            raise ValueError("Mask and branch offset have different shape.")

        # Set branch offsets
        self.branch_offsets = branch_offsets
        # Calculate node positions
        if branch_offsets is not None:
            self.node_xy = self.interpolate(branch_offsets)

        # Set which of the nodes are present
        if (mask is None) and (branch_offsets is not None):
            self.mask = np.full(branch_offsets.shape, False)
        else:
            self.mask = mask

    def generate_nodes(
        self,
        mesh1d_edge_length: float,
        structure_chainage: Optional[List[float]] = None,
        max_dist_to_struc: Optional[float] = None,
    ):
        """Generate the branch offsets and the nodes.

        Args:
            mesh1d_edge_length (float): The edge length of the 1d mesh.
            structure_chainage (Optional[List[float]], optional): A list with the structure chainages. If not specified, calculation will not take it into account. Defaults to None.
            max_dist_to_struc (Optional[float], optional): The maximum distance from a node to a structure. If not specified, calculation will not take it into account. Defaults to None.

        Raises:
            ValueError: Raised when any of the structure offsets, if specified, is smaller than zero.
            ValueError: Raised when any of the structure offsets, if specified, is greater than the branch length.
        """
        # Generate offsets
        self.branch_offsets = self._generate_offsets(
            mesh1d_edge_length, structure_chainage, max_dist_to_struc
        )
        # Calculate node positions
        self.node_xy = self.interpolate(self.branch_offsets)
        # Add mask (all False)
        self.mask = np.full(self.branch_offsets.shape, False)

    def _generate_offsets(
        self,
        mesh1d_edge_length: float,
        structure_offsets: Optional[List[float]] = None,
        max_dist_to_struc: Optional[float] = None,
    ) -> np.ndarray:
        """Generate the branch offsets.

        Args:
            mesh1d_edge_length (float): The edge length of the 1d mesh.
            structure_chainage (Optional[List[float]], optional): A list with the structure chainages. If not specified, calculation will not take it into account. Defaults to None.
            max_dist_to_struc (Optional[float], optional): The maximum distance from a node to a structure. If not specified, calculation will not take it into account. Defaults to None.

        Raises:
            ValueError: Raised when any of the structure offsets, if specified, is smaller than zero.
            ValueError: Raised when any of the structure offsets, if specified, is greater than the branch length.

        Returns:
            np.ndarray: The generated branch offsets.
        """
        # Generate initial offsets
        anchor_pts = [0.0, self.length]
        offsets = self._generate_1d_spacing(anchor_pts, mesh1d_edge_length)

        if structure_offsets is None:
            return offsets

        # Check the limits
        if (excess := min(structure_offsets)) < 0.0 or (
            excess := max(structure_offsets)
        ) > self.length:
            raise ValueError(
                f"Distance {excess} is outside the branch range (0.0 - {self.length})."
            )

        # Merge limits with start and end of branch
        limits = [-1e-3] + list(sorted(structure_offsets)) + [self.length + 1e-3]

        # if requested, check if the calculation point are close enough to the structures
        if max_dist_to_struc is not None:
            limits = self._generate_extended_limits(max_dist_to_struc, limits)

        offsets = self._add_nodes_to_segments(
            offsets, anchor_pts, limits, mesh1d_edge_length
        )

        return offsets

    def _generate_extended_limits(
        self, max_dist_to_struc: float, limits: List[float]
    ) -> List[float]:
        """Generate extended limits by taking into account the maximum distance to a structure.

        Args:
            max_dist_to_struc (float): The maximum distance from a node to a structure.
            limits (List[float]): The limits.

        Returns:
            List[float]: A list with the updated limits.
        """

        additional = []

        # Skip the first and the last, these are no structures
        for i in range(1, len(limits) - 1):
            # if the distance between two limits is large than twice the max distance to structure,
            # the mesh point will be too far away. Add a limit on the minimum of half the length and
            # two times the max distance
            dist_to_prev_limit = limits[i] - (
                max(additional[-1], limits[i - 1]) if any(additional) else limits[i - 1]
            )
            if dist_to_prev_limit > 2 * max_dist_to_struc:
                additional.append(
                    limits[i] - min(2 * max_dist_to_struc, dist_to_prev_limit / 2)
                )

            dist_to_next_limit = limits[i + 1] - limits[i]
            if dist_to_next_limit > 2 * max_dist_to_struc:
                additional.append(
                    limits[i] + min(2 * max_dist_to_struc, dist_to_next_limit / 2)
                )

        # Join the limits
        return sorted(limits + additional)

    def _add_nodes_to_segments(
        self,
        offsets: np.ndarray,
        anchor_pts: List[float],
        limits: List[float],
        mesh1d_edge_length: float,
    ) -> np.ndarray:
        """Add nodes to segments that are missing a mesh node.

        Args:
            offsets (np.ndarray): The branch offsets.
            anchor_pts (List[float]): The anchor points.
            limits (List[float]): The limits.
            mesh1d_edge_length (float): The edge length of the 1d mesh.

        Returns:
            np.ndarray: The array with branch offsets.
        """
        # Get upper and lower limits
        upper_limits = limits[1:]
        lower_limits = limits[:-1]

        def in_range():
            return [
                ((offsets > lower) & (offsets < upper)).any()
                for lower, upper in zip(lower_limits, upper_limits)
            ]

        # Determine the segments that are missing a mesh node
        # Anchor points are added on these segments, such that they will get a mesh node
        nodes_in_range = in_range()

        while not all(nodes_in_range):
            # Get the index of the first segment without grid point
            i = nodes_in_range.index(False)

            # Add it to the anchor pts
            anchor_pts.append((lower_limits[i] + upper_limits[i]) / 2.0)
            anchor_pts = sorted(anchor_pts)

            # Generate new offsets
            offsets = self._generate_1d_spacing(anchor_pts, mesh1d_edge_length)

            # Determine the segments that are missing a grid point
            nodes_in_range = in_range()

        if len(anchor_pts) > 2:
            logger.info(
                f"Added 1d mesh nodes on branch at: {anchor_pts}, due to the structures at {limits}."
            )

        return offsets

    @staticmethod
    def _generate_1d_spacing(
        anchor_pts: List[float], mesh1d_edge_length: float
    ) -> np.ndarray:
        """
        Generates 1d distances, called by function generate offsets
        """
        offsets = []
        # Loop through anchor point pairs
        for i in range(len(anchor_pts) - 1):
            # Determine section length between anchor point
            section_length = anchor_pts[i + 1] - anchor_pts[i]
            if section_length <= 0.0:
                raise ValueError("Section length must be larger than 0.0")
            # Determine number of nodes
            nnodes = max(2, int(round(section_length / mesh1d_edge_length) + 1)) - 1
            # Add nodes
            offsets.extend(
                np.linspace(
                    anchor_pts[i], anchor_pts[i + 1], nnodes, endpoint=False
                ).tolist()
            )
        # Add last node
        offsets.append(anchor_pts[-1])

        return np.asarray(offsets)

    def interpolate(self, distance: npt.ArrayLike) -> np.ndarray:
        """Interpolate coordinates along branch by length

        Args:
            distance (npt.ArrayLike): Length
        """
        intpcoords = np.stack(
            [
                np.interp(distance, self._distance, self._x_coordinates),
                np.interp(distance, self._distance, self._y_coordinates),
            ],
            axis=1,
        )

        return intpcoords


class Link1d2d(BaseModel):
    """Link1d2d defines the 1D2D Links of a model network.

    Attributes:
        meshkernel (Optional[mk.MeshKernel]):
            The MeshKernel used to interact with this Link1d2d
        link1d2d_id (np.ndarray):
            The id of this Link1d2d
        link1d2d_long_name (np.ndarray):
            The long name of this Link1d2d
        link1d2d_contact_type (np.ndarray):
            The contact type of this Link1d2d
        link1d2d (np.ndarray):
            The underlying data object of this Link1d2d
    """

    meshkernel: mk.MeshKernel = Field(default_factory=mk.MeshKernel)

    link1d2d_id: np.ndarray = Field(default_factory=lambda: np.empty(0, object))
    link1d2d_long_name: np.ndarray = Field(default_factory=lambda: np.empty(0, object))
    link1d2d_contact_type: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.int32)
    )
    link1d2d: np.ndarray = Field(default_factory=lambda: np.empty((0, 2), np.int32))

    def is_empty(self) -> bool:
        """Whether this Link1d2d is currently empty.

        Returns:
            bool: True if the Link1d2d is currently empty; False otherwise.
        """
        return self.link1d2d.size == 0

    def read_file(self, file_path: Path) -> None:
        """Read the Link1d2d data from the specified netCDF file at file_path into this

        Args:
            file_path (Path): Path to the netCDF file.
        """

        reader = UgridReader(file_path)
        reader.read_link1d2d(self)

    def clear(self) -> None:
        """Remove all saved links from the links administration"""
        self.link1d2d_id = np.empty(0, object)
        self.link1d2d_long_name = np.empty(0, object)
        self.link1d2d_contact_type = np.empty(0, np.int32)
        self.link1d2d = np.empty((0, 2), np.int32)
        # The meshkernel object needs to be resetted
        self.meshkernel._deallocate_state()
        self.meshkernel._allocate_state(self.meshkernel.get_projection())
        self.meshkernel.contacts_get()

    def _process(self) -> None:
        """
        Get links from meshkernel and add to the array with link administration
        """
        contacts = self.meshkernel.contacts_get()

        self.link1d2d = np.append(
            self.link1d2d,
            np.stack([contacts.mesh1d_indices, contacts.mesh2d_indices], axis=1),
            axis=0,
        )
        self.link1d2d_contact_type = np.append(
            self.link1d2d_contact_type, np.full(contacts.mesh1d_indices.size, 3)
        )
        self.link1d2d_id = np.append(
            self.link1d2d_id,
            np.array([f"{n1d:d}_{f2d:d}" for n1d, f2d in self.link1d2d]),
        )
        self.link1d2d_long_name = np.append(
            self.link1d2d_long_name,
            np.array([f"{n1d:d}_{f2d:d}" for n1d, f2d in self.link1d2d]),
        )

    def _link_from_1d_to_2d(
        self, node_mask: np.ndarray, polygon: mk.GeometryList = None
    ):
        """Connect 1d nodes to 2d face circumcenters. A list of branchid's can be given
        to indicate where the 1d-side of the connections should be made. A polygon can
        be given to indicate where the 2d-side of the connections should be made.

        Note that the links are added to the already existing links. To remove these, use the method "clear".

        Args:
            node_mask (np.ndarray): Array indicating what 1d nodes should be connected. Defaults to None.
            polygon (mk.GeometryList): Coordinates of the area within which the 2d side of the links are connected.
        """

        # Computes Mesh1d-Mesh2d contacts, where each single Mesh1d node is connected to one Mesh2d face circumcenter.
        # The boundary nodes of Mesh1d (those sharing only one Mesh1d edge) are not connected to any Mesh2d face.
        self.meshkernel.contacts_compute_single(
            node_mask=node_mask, polygons=polygon, projection_factor=1.0
        )
        self._process()

        # Note that the function "contacts_compute_multiple" also computes the connections, but does not take into account
        # a bounding polygon or the end points of the 1d mesh.

    def _link_from_2d_to_1d_embedded(
        self, node_mask: np.ndarray, points: mk.GeometryList
    ):
        """"""
        self.meshkernel.contacts_compute_with_points(node_mask=node_mask, points=points)
        self._process()

    def _link_from_2d_to_1d_lateral(
        self,
        node_mask: np.ndarray,
        # boundary_face_xy: np.ndarray,
        polygon: mk.GeometryList = None,
        search_radius: float = None,
    ):
        # TODO: Missing value double for search radius?

        # Computes Mesh1d-Mesh2d contacts, where Mesh1d nodes are connected to the closest Mesh2d faces at the boundary
        self.meshkernel.contacts_compute_boundary(
            node_mask=node_mask, polygons=polygon, search_radius=search_radius
        )
        self._process()


class Mesh1d(BaseModel):
    """"""

    meshkernel: mk.MeshKernel = Field(default_factory=mk.MeshKernel)

    branches: Dict[str, Branch] = {}

    network1d_node_id: np.ndarray = Field(default_factory=lambda: np.empty(0, object))
    network1d_node_long_name: np.ndarray = Field(
        default_factory=lambda: np.empty(0, object)
    )
    network1d_node_x: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    network1d_node_y: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    network1d_branch_id: np.ndarray = Field(default_factory=lambda: np.empty(0, object))
    network1d_branch_long_name: np.ndarray = Field(
        default_factory=lambda: np.empty(0, object)
    )
    network1d_branch_length: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.double)
    )
    network1d_branch_order: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.int32)
    )
    network1d_edge_nodes: np.ndarray = Field(
        default_factory=lambda: np.empty((0, 2), np.int32)
    )
    # TODO: sync with node_x/node_y/edge_nodes with meshkernel: https://github.com/Deltares/HYDROLIB-core/issues/576
    network1d_geom_x: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    network1d_geom_y: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    network1d_part_node_count: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.int32)
    )

    mesh1d_node_x: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    mesh1d_node_y: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    mesh1d_node_id: np.ndarray = Field(default_factory=lambda: np.empty(0, object))
    mesh1d_node_long_name: np.ndarray = Field(
        default_factory=lambda: np.empty(0, object)
    )
    mesh1d_node_branch_id: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.int32)
    )
    mesh1d_node_branch_offset: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.double)
    )

    mesh1d_edge_nodes: np.ndarray = Field(
        default_factory=lambda: np.empty((0, 2), np.int32)
    )
    mesh1d_edge_x: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    mesh1d_edge_y: np.ndarray = Field(default_factory=lambda: np.empty(0, np.double))
    mesh1d_edge_branch_id: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.int32)
    )
    mesh1d_edge_branch_offset: np.ndarray = Field(
        default_factory=lambda: np.empty(0, np.double)
    )

    def is_empty(self) -> bool:
        return self.mesh1d_node_x.size == 0

    def _get_mesh1d(self) -> mk.Mesh1d:
        """Return mesh1d from meshkernel. Note that the meshkernel.Mesh1d instance
        does not contain all mesh attributes that are contained in this class"""
        return self.meshkernel.mesh1d_get()

    def _set_mesh1d(self) -> None:
        self.meshkernel.mesh1d_set(
            mk.Mesh1d(
                node_x=self.mesh1d_node_x.astype(np.float64),
                node_y=self.mesh1d_node_y.astype(np.float64),
                edge_nodes=self.mesh1d_edge_nodes.ravel().astype(np.int32),
            )
        )

    def _process_network1d(self) -> None:
        """
        Determine x, y locations of mesh1d nodes based on the network1d
        """
        # Create a list of coordinates to create the branches from
        ngeom = list(zip(self.network1d_geom_x, self.network1d_geom_y))

        self.branches.clear()

        for i, (name, nnodes) in enumerate(
            zip(self.network1d_branch_id, self.network1d_part_node_count)
        ):

            # Create network branch
            # Get geometry of branch from network geometry
            geometry = np.array([ngeom.pop(0) for _ in range(nnodes)])
            # Get branch offsets
            idx = self.mesh1d_node_branch_id == i
            branch_offsets = self.mesh1d_node_branch_offset[idx]
            mask = np.full(branch_offsets.shape, False)

            # Determine if a start or end coordinate needs to be added for constructing a complete branch
            # As nodes are re-used, the last and first branch_offsets are often missing. However, they are still used
            # for determining the length along the discretized branch.
            if branch_offsets.size == 0 or not np.isclose(branch_offsets[0], 0.0):
                branch_offsets = np.concatenate([[0], branch_offsets])
                mask = np.concatenate([[True], mask])
            length = np.hypot(*np.diff(geometry, axis=0).T).sum()
            if not np.isclose(branch_offsets[-1], length):
                branch_offsets = np.concatenate([branch_offsets, [length]])
                mask = np.concatenate([mask, [True]])

            # Create instance of branch object and add to dictionary
            geo_branch = Branch(geometry, branch_offsets=branch_offsets, mask=mask)
            self.branches[name.strip()] = geo_branch

        # Convert list with all coordinates (except the appended ones for the schematized branches) to arrays
        node_x, node_y = np.vstack(
            [branch.node_xy[~branch.mask] for branch in self.branches.values()]
        ).T

        # Add to variables
        self.mesh1d_node_x = node_x
        self.mesh1d_node_y = node_y

        # Calculate edge coordinates
        edge_x, edge_y = np.vstack(
            [
                branch.interpolate(
                    self.mesh1d_edge_branch_offset[self.mesh1d_edge_branch_id == i]
                )
                for i, branch in enumerate(self.branches.values())
            ]
        ).T

        # Add to variables
        self.mesh1d_edge_x = edge_x
        self.mesh1d_edge_y = edge_y

    def _network1d_node_position(self, x: float, y: float) -> Union[np.int32, None]:
        """Determine the position (index) of a x, y coordinate in the network nodes

        Args:
            x (float): x-coordinate
            y (float): y-coordinate

        Returns:
            Union[np.int32, None]: The index of the coordinate. None if not found
        """
        return self._node_position(self.network1d_node_x, self.network1d_node_y, x, y)

    def _mesh1d_node_position(self, x: float, y: float) -> Union[np.int32, None]:
        """Determine the position (index) of a x, y coordinate in the mesh nodes

        Args:
            x (float): x-coordinate
            y (float): y-coordinate

        Returns:
            Union[np.int32, None]: The index of the coordinate. None if not found
        """
        return self._node_position(self.mesh1d_node_x, self.mesh1d_node_y, x, y)

    def _node_position(
        self, arrx: np.ndarray, arry: np.ndarray, x: float, y: float
    ) -> Union[np.int32, None]:
        """Determine the position (index) of a x, y coordinate in a given x and y array

        Args:
            arrx (np.ndarray): x-coordinates in which the position is sought
            arry (np.ndarray): y-coordiantes in which the position is sought
            x (float): x-coordinate to be sought
            y (float): y-coordinate to be sought

        Raises:
            ValueError: If multiple positions are found for the coordinate

        Returns:
            Union[np.int32, None]: The index of the coordinate. None if not found
        """
        pos = np.where(np.isclose(arrx, x, rtol=0.0) & np.isclose(arry, y, rtol=0.0))[0]
        if pos.size == 0:
            return None
        elif pos.size == 1:
            return np.int32(pos[0])
        else:
            # Find the nearest
            distance = np.hypot(arrx[pos] - x, arry[pos] - y)
            if np.unique(distance).size == 1:
                raise ValueError("Multiple nodes were found at the same position.")
            else:
                return np.int32(pos[np.argmin(distance)])

    def _add_branch(
        self,
        branch: Branch,
        name: str = None,
        branch_order: int = -1,
        long_name: str = None,
        force_midpoint: bool = True,
    ):
        """Add the branch to mesh1d

        Args:
            branch (Branch): branch to add to the mesh1d
            name (str): id of the branch
            branch_order (int): interpolation order of the branch
            long_name (str): long name of the branch
            force_midpoint(bool): argument to control if a midpoint will be forced on the branch, use False for pipes

        Returns:
            Str: name of the branch.
        """

        # Check if branch had coordinate discretization
        if branch.branch_offsets.size == 0:
            raise ValueError(
                'Branch has no mesh discretization. Use the function "generate_nodes" solve generate a 1d mesh on the branch.'
            )

        if name in self.network1d_branch_id:
            raise KeyError(f'The branch name "{name}" is already used.')
        if long_name in self.network1d_branch_long_name:
            raise KeyError(f'The branch long name "{long_name}" is already used.')

        branch_nr = len(self.network1d_branch_id)
        if name is None:
            name = f"br{branch_nr:05d}"
        if long_name is None:
            long_name = name

        self.branches[name] = branch

        # Add branch administration
        self.network1d_branch_order = np.append(
            self.network1d_branch_order, branch_order
        )
        self.network1d_branch_length = np.append(
            self.network1d_branch_length, branch.length
        )
        self.network1d_branch_id = np.append(self.network1d_branch_id, name)
        self.network1d_branch_long_name = np.append(
            self.network1d_branch_long_name, long_name
        )

        # Add branch geometry coordinates
        self.network1d_part_node_count = np.append(
            self.network1d_part_node_count, len(branch.geometry)
        )
        self.network1d_geom_x = np.append(self.network1d_geom_x, branch._x_coordinates)
        self.network1d_geom_y = np.append(self.network1d_geom_y, branch._y_coordinates)

        # Network edge node administration
        # -------------------------------

        first_point = branch.geometry[0]
        last_point = branch.geometry[-1]

        # Get offsets from dictionary
        offsets = branch.branch_offsets[:]
        # The number of links on the branch
        nlinks = len(offsets) - 1

        # Check if the first and last point of the branch are already in the set
        first_present = self._network1d_node_position(*first_point) is not None
        if first_present:
            # If present, remove from branch offsets
            offsets = offsets[1:]
            branch.mask[0] = True
        else:
            # If not present, add to network nodes
            self.network1d_node_x = np.append(self.network1d_node_x, first_point[0])
            self.network1d_node_y = np.append(self.network1d_node_y, first_point[1])

            self.network1d_node_id = np.append(
                self.network1d_node_id, "{:.6f}_{:.6f}".format(*first_point)
            )
            self.network1d_node_long_name = np.append(
                self.network1d_node_long_name, "x={:.6f}_y={:.6f}".format(*first_point)
            )

        last_present = self._network1d_node_position(*last_point) is not None
        if last_present:
            # If present, remove from branch offsets
            offsets = offsets[:-1]
            branch.mask[-1] = True
        else:
            # If not present, add to network nodes
            self.network1d_node_x = np.append(self.network1d_node_x, last_point[0])
            self.network1d_node_y = np.append(self.network1d_node_y, last_point[1])

            self.network1d_node_id = np.append(
                self.network1d_node_id, "{:.6f}_{:.6f}".format(*last_point)
            )
            self.network1d_node_long_name = np.append(
                self.network1d_node_long_name, "x={:.6f}_y={:.6f}".format(*last_point)
            )

        # If no points remain, add an extra halfway: each branch should have at least 1 node
        # Adjust the branch object as well, by adding the extra point
        if len(offsets) == 0 and force_midpoint:
            # Add extra offset
            extra_offset = branch.length / 2.0
            offsets = np.array([extra_offset])
            nlinks += 1
            # Adjust branch object
            branch.branch_offsets = np.insert(branch.branch_offsets, 1, extra_offset)
            branch.node_xy = np.insert(
                branch.node_xy, 1, branch.interpolate(offsets), axis=0
            )
            branch.mask = np.insert(branch.mask, 1, False)

        # Get the index of the first and last node, add as edge_nodes
        i_from = self._network1d_node_position(first_point[0], first_point[1])
        i_to = self._network1d_node_position(last_point[0], last_point[1])
        if i_from == i_to:
            raise ValueError(
                "Start and end node are the same. Ring geometries are not accepted."
            )

        self.network1d_edge_nodes = np.append(
            self.network1d_edge_nodes,
            np.array([[i_from, i_to]], dtype=np.int32),
            axis=0,
        )

        # Mesh1d edge node administration

        # -------------------------------
        # First determine the start index. This is equal to the number of already present points
        start_index = len(self.mesh1d_node_branch_id)
        # For each link, create a new edge node connection
        # If the first node is already present, subtract 1, since the first number will be substitud with the present node
        if first_present:
            start_index -= 1
        new_edge_nodes = (
            np.stack([np.arange(nlinks), np.arange(nlinks) + 1], axis=1) + start_index
        ).astype(np.int32)

        # If the first node is present, change the first point of the first edge to the existing point
        if first_present:
            new_edge_nodes[0, 0] = self._mesh1d_node_position(*first_point)
        # If the last node is present, change the last point of the last edge too
        if last_present:
            new_edge_nodes[-1, 1] = self._mesh1d_node_position(*last_point)

        # Add to variables
        self.mesh1d_node_x = np.append(
            self.mesh1d_node_x, branch.node_xy[~branch.mask, 0]
        )
        self.mesh1d_node_y = np.append(
            self.mesh1d_node_y, branch.node_xy[~branch.mask, 1]
        )

        # Add to edge_nodes
        self.mesh1d_edge_nodes = np.append(
            self.mesh1d_edge_nodes, new_edge_nodes, axis=0
        )
        edge_coords = np.stack([self.mesh1d_node_x, self.mesh1d_node_y], axis=1)[
            new_edge_nodes
        ].mean(1)
        edge_offsets = (branch.branch_offsets[:-1] + branch.branch_offsets[1:]) / 2

        self.mesh1d_edge_branch_id = np.append(
            self.mesh1d_edge_branch_id, np.full(len(edge_coords), branch_nr)
        )
        self.mesh1d_edge_branch_offset = np.append(
            self.mesh1d_edge_branch_offset, edge_offsets
        )

        self.mesh1d_edge_x = np.append(self.mesh1d_edge_x, edge_coords[:, 0])
        self.mesh1d_edge_y = np.append(self.mesh1d_edge_y, edge_coords[:, 1])

        # Update names of nodes
        mesh_point_names = np.array(
            [f"{name}_{offset:.2f}" for offset in offsets], dtype=object
        )
        self.mesh1d_node_id = np.append(self.mesh1d_node_id, mesh_point_names)
        self.mesh1d_node_long_name = np.append(
            self.mesh1d_node_long_name, mesh_point_names
        )

        # Add mesh1d nodes
        self.mesh1d_node_branch_id = np.append(
            self.mesh1d_node_branch_id, np.full(len(offsets), branch_nr)
        )
        self.mesh1d_node_branch_offset = np.append(
            self.mesh1d_node_branch_offset, offsets
        )
        return name

    def get_node_mask(self, branchids: List[str] = None):
        """Get node mask, give a mask with True for each node that is in the given branchid list"""

        mask = np.full(self.mesh1d_node_id.shape, False, dtype=bool)
        if branchids is None:
            mask[:] = True
            return mask

        # Get number (index) of given branches
        idx = np.where(np.isin(self.network1d_branch_id, branchids))[0]
        if idx.size == 0:
            raise KeyError("No branches corresponding to the given keys were found.")

        mask[np.isin(self.mesh1d_node_branch_id, idx)] = True

        return mask


class Network:
    def __init__(self, is_geographic: bool = False) -> None:
        if not is_geographic:
            projection = mk.ProjectionType.CARTESIAN
        else:
            projection = mk.ProjectionType.SPHERICAL

        self.meshkernel = mk.MeshKernel(projection=projection)
        self._mesh1d = Mesh1d(meshkernel=self.meshkernel)
        self._mesh2d = Mesh2d(meshkernel=self.meshkernel)
        self._link1d2d = Link1d2d(meshkernel=self.meshkernel)

        # Spatial index (rtree)
        # self._idx = index.Index()

    @classmethod
    def from_file(cls, file_path: Path) -> Network:
        """Read network from file. This classmethod checks what mesh components (mesh1d & network1d, mesh2d, link1d2d) are
        present, and loads them one by one.

        Args:
            file_path (Path): path to netcdf file with network data

        Returns:
            Network: The instance of the class itself that is returned
        """

        network = cls()
        ds = nc.Dataset(file_path)  # type: ignore[import]

        reader = UgridReader(file_path)

        reader.read_mesh1d_network1d(network._mesh1d)
        reader.read_mesh2d(network._mesh2d)
        reader.read_link1d2d(network._link1d2d)

        ds.close()

        return network

    def to_file(self, file: Path) -> None:
        """Write network to file

        Args:
            file (Path): File where _net.nc is written to.
        """

        writer = UgridWriter()
        writer.write(self, file)

    @property
    def is_geographic(self) -> bool:
        """Whether or not this network has a geographic projection.

        Returns:
            bool: True if this network is geographic; otherwise, False.
        """
        projection = self.meshkernel.get_projection()
        if projection == mk.ProjectionType.CARTESIAN:
            return False
        else:
            return True

    def link1d2d_from_1d_to_2d(
        self, branchids: List[str] = None, polygon: GeometryList = None
    ) -> None:
        self._mesh1d._set_mesh1d()

        node_mask = self._mesh1d.get_node_mask(branchids)
        if polygon is None:
            polygon = self.meshkernel.mesh2d_get_mesh_boundaries_as_polygons()

        self._link1d2d._link_from_1d_to_2d(node_mask, polygon=polygon)

    def mesh2d_create_rectilinear_within_extent(
        self, extent: tuple, dx: float, dy: float
    ) -> None:
        self._mesh2d.create_rectilinear(extent=extent, dx=dx, dy=dy)

    def mesh2d_create_triangular_within_polygon(self, polygon: mk.GeometryList) -> None:
        """Create triangular grid within GeometryList object. Calls _mesh2d.create_triangular
        directly, but is easier accessible for users.

        Args:
            polygon (mk.GeometryList): GeometryList representing a polygon within which the mesh is generated.
        """
        self._mesh2d.create_triangular(geometry_list=polygon)

    def mesh2d_clip_mesh(
        self,
        geometrylist: mk.GeometryList,
        deletemeshoption: mk.DeleteMeshOption = mk.DeleteMeshOption.INSIDE_NOT_INTERSECTED,
        inside=True,
    ) -> None:
        self._mesh2d.clip(
            geometrylist=geometrylist,
            deletemeshoption=deletemeshoption,
            inside=inside,
        )

    def mesh2d_refine_mesh(self, polygon: mk.GeometryList, level: int = 1) -> None:
        self._mesh2d.refine(polygon=polygon, level=level)

    def mesh1d_add_branch(
        self,
        branch: Branch,
        name: str = None,
        branch_order: int = -1,
        long_name: str = None,
        force_midpoint: bool = True,
    ) -> None:
        name = self._mesh1d._add_branch(
            branch=branch,
            name=name,
            branch_order=branch_order,
            long_name=long_name,
            force_midpoint=force_midpoint,
        )
        self._mesh1d._set_mesh1d()
        return name

    def plot(self, ax=None):
        """Create a plot of the 1d2d links and edges within this network.

        Args:
            ax (matplotlib.pyplot.Axes, optional): The axes where to plot the edges. Defaults to None.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots()
        mesh2d_output = self._mesh2d.get_mesh2d()
        mesh1d_output = self._mesh1d._get_mesh1d()
        links_output = self._link1d2d.meshkernel.contacts_get()
        mesh2d_output.plot_edges(ax=ax, color="r")
        mesh1d_output.plot_edges(ax=ax, color="g")
        links_output.plot_edges(
            ax=ax, mesh1d=mesh1d_output, mesh2d=mesh2d_output, color="k"
        )


class NetworkModel(ParsableFileModel):
    """Network model representation."""

    network: Network = Field(default_factory=Network)

    def _post_init_load(self) -> None:
        """
        Load the network file if the filepath exists relative to the
        current FileLoadContext.
        """
        super()._post_init_load()

        if self.filepath is None:
            return

        with file_load_context() as context:
            network_path = context.resolve(self.filepath)

            if network_path.is_file():
                self.network = Network.from_file(network_path)

    @property
    def _mesh1d(self):
        return self.network._mesh1d

    @property
    def _mesh2d(self):
        return self.network._mesh2d

    @property
    def _link1d2d(self):
        return self.network._link1d2d

    @classmethod
    def _ext(cls) -> str:
        return ".nc"

    @classmethod
    def _filename(cls) -> str:
        return "network"

    def _save(self, save_settings: ModelSaveSettings):
        with file_load_context() as context:
            write_path = context.resolve(self.filepath)  # type: ignore[arg-type]

            write_path.parent.mkdir(parents=True, exist_ok=True)
            self.network.to_file(write_path)

    def _export(self, folder: Path) -> None:
        filename = Path(self.filepath.name) if self.filepath else self._generate_name()
        self.filepath = folder / filename
        folder.mkdir(parents=True, exist_ok=True)
        self.network.to_file(self.filepath)

    def _parse(self, _):
        return {}

    @classmethod
    def _get_serializer(cls):
        # Unused, but requires abstract implementation
        pass

    @classmethod
    def _get_parser(cls):
        # Unused, but requires abstract implementation
        pass

    @property
    def plot(self):
        return self.network.plot



from itertools import chain, count, repeat
from pathlib import Path
from typing import Any, Iterable, Optional, Sequence

from hydrolib.core.basemodel import BaseModel, SerializerConfig
from hydrolib.core.dflowfm.ini.io_models import (
    CommentBlock,
    ContentElement,
    Datablock,
    DatablockRow,
    Document,
    Property,
    Section,
)
from hydrolib.core.utils import str_is_empty_or_none


class INISerializerConfig(SerializerConfig):
    """SerializerConfig defines the configuration options of the Serializer

    Attributes:
        section_indent (int):
            The number of spaces with which whole sections should be indented.
            Defaults to 0.
        property_indent (int):
            The number of spaces with which properties should be indented relative to
            the section header (i.e. the full indent equals the section_indent plus
            property_indent). Defaults to 4.
        datablock_indent (int):
            The number of spaces with which datablock rows are indented relative to
            the section header (i.e. the full indent equals the section_indent plus
            datablock_indent). Defaults to 8.
        datablock_spacing (int):
            The number of spaces between datablock columns. Note that there might be
            additional offset to ensure . is lined out. Defaults to 2.
        comment_delimiter (str):
            The character used to delimit comments. Defaults to '#'.
        skip_empty_properties (bool):
            Whether or not to skip properties with a value that is empty or None. Defaults to True.
    """

    section_indent: int = 0
    property_indent: int = 0
    datablock_indent: int = 8
    datablock_spacing: int = 2
    comment_delimiter: str = "#"
    skip_empty_properties: bool = True

    @property
    def total_property_indent(self) -> int:
        """The combined property indentation, i.e. section_indent + property_indent"""
        return self.section_indent + self.property_indent

    @property
    def total_datablock_indent(self) -> int:
        """The combined datablock indentation, i.e. section_indent + datablock_indent"""
        return self.section_indent + self.datablock_indent


class DataBlockINIBasedSerializerConfig(INISerializerConfig):
    """Class that holds the configuration settings for INI files with data blocks serialization."""

    float_format_datablock: str = ""
    """str: The string format that will be used for float serialization of the datablock. If empty, the original number will be serialized. Defaults to an empty string.
    
    Examples:
        Input value = 123.456

        Format    | Output          | Description
        -------------------------------------------------------------------------------------------------------------------------------------
        ".0f"     | 123             | Format float with 0 decimal places.
        "f"       | 123.456000      | Format float with default (=6) decimal places.
        ".2f"     | 123.46          | Format float with 2 decimal places.
        "+.1f"    | +123.5          | Format float with 1 decimal place with a + or  sign.
        "e"       | 1.234560e+02    | Format scientific notation with the letter 'e' with default (=6) decimal places.
        "E"       | 1.234560E+02    | Format scientific notation with the letter 'E' with default (=6) decimal places.
        ".3e"     | 1.235e+02       | Format scientific notation with the letter 'e' with 3 decimal places.
        "<15"     | 123.456         | Left aligned in space with width 15
        "^15.0f"  |       123       | Center aligned in space with width 15 with 0 decimal places.
        ">15.1e"  |         1.2e+02 | Right aligned in space with width 15 with scientific notation with 1 decimal place.
        "*>15.1f" | **********123.5 | Right aligned in space with width 15 with 1 decimal place and fill empty space with *
        "%"       | 12345.600000%   | Format percentage with default (=6) decimal places.     
        ".3%"     | 12345.600%      | Format percentage with 3 decimal places.  

        More information: https://docs.python.org/3/library/string.html#format-specification-mini-language
    """


class MaxLengths(BaseModel):
    """MaxLengths defines the maxmimum lengths of the parts of a section

    Attributes:
        key (int):
            The maximum length of all the keys of the properties within a section.
            If no properties are present it should be 0.
        value (int):
            The maximum length of all the non None values of the properties within a
            section. If no properties are present, or all values are None, it should
            be 0.
        datablock (Optional[Sequence[int]]):
            The maximum length of the values of each column of the Datablock.
            If no datablock is present it defaults to None.
    """

    key: int
    value: int
    datablock: Optional[Sequence[int]] = None

    @classmethod
    def from_section(cls, section: Section) -> "MaxLengths":
        """Generate a MaxLengths instance from the given Section

        Args:
            section (Section): The section of which the MaxLengths are calculated

        Returns:
            MaxLengths: The MaxLengths corresponding with the provided section
        """
        properties = list(p for p in section.content if isinstance(p, Property))

        keys = (prop.key for prop in properties)
        values = (prop.value for prop in properties if prop.value is not None)

        max_key_length = max((len(k) for k in keys), default=0)
        max_value_length = max((len(v) for v in values), default=0)
        max_datablock_lengths = MaxLengths._of_datablock(section.datablock)

        return cls(
            key=max_key_length,
            value=max_value_length,
            datablock=max_datablock_lengths,
        )

    @staticmethod
    def _of_datablock(datablock: Optional[Datablock]) -> Optional[Sequence[int]]:
        if datablock is None or len(datablock) < 1:
            return None

        datablock_columns = map(list, zip(*datablock))
        datablock_column_lengths = (map(len, column) for column in datablock_columns)  # type: ignore
        max_lengths = (max(column) for column in datablock_column_lengths)

        return tuple(max_lengths)


Lines = Iterable[str]


def _serialize_comment_block(
    block: CommentBlock,
    delimiter: str = "#",
    indent_size: int = 0,
) -> Lines:
    indent = " " * indent_size
    return (f"{indent}{delimiter} {l}" for l in block.lines)


def _get_offset_whitespace(key: Optional[str], max_length: int) -> str:
    key_length = len(key) if key is not None else 0
    return " " * max(max_length - key_length, 0)


class SectionSerializer:
    """SectionSerializer provides the serialize method to serialize a Section

    The entrypoint of this method is the serialize method, which will construct
    an actual instance and serializes the Section with it.
    """

    def __init__(self, config: INISerializerConfig, max_length: MaxLengths):
        """Create a new SectionSerializer

        Args:
            config (SerializerConfig): The config describing the serialization options
            max_length (MaxLengths): The max lengths of the section being serialized
        """
        self._config = config
        self._max_length = max_length

    @classmethod
    def serialize(cls, section: Section, config: INISerializerConfig) -> Lines:
        """Serialize the provided section with the given config

        Args:
            section (Section): The section to serialize
            config (SerializerConfig): The config describing the serialization options

        Returns:
            Lines: The iterable lines of the serialized section
        """
        serializer = cls(config, MaxLengths.from_section(section))
        return serializer._serialize_section(section)

    @property
    def config(self) -> INISerializerConfig:
        """The SerializerConfig used while serializing the section."""
        return self._config

    @property
    def max_length(self) -> MaxLengths:
        """The MaxLengths of the Section being serialized by this SectionSerializer."""
        return self._max_length

    def _serialize_section(self, section: Section) -> Lines:
        header_iterable = self._serialize_section_header(section.header)
        properties = self._serialize_content(section.content)
        datablock = self._serialize_datablock(section.datablock)

        return chain(header_iterable, properties, datablock)

    def _serialize_section_header(self, section_header: str) -> Lines:
        indent = " " * (self.config.section_indent)
        yield f"{indent}[{section_header}]"

    def _serialize_content(self, content: Iterable[ContentElement]) -> Lines:
        elements = (self._serialize_content_element(elem) for elem in content)
        return chain.from_iterable(elements)

    def _serialize_content_element(self, elem: ContentElement) -> Lines:
        if isinstance(elem, Property):
            return self._serialize_property(elem)
        else:
            indent = self.config.total_property_indent
            delimiter = self.config.comment_delimiter
            return _serialize_comment_block(elem, delimiter, indent)

    def _serialize_property(self, property: Property) -> Lines:
        if self.config.skip_empty_properties and str_is_empty_or_none(property.value):
            return

        indent = " " * (self._config.total_property_indent)
        key_ws = _get_offset_whitespace(property.key, self.max_length.key)
        key = f"{property.key}{key_ws} = "

        value_ws = _get_offset_whitespace(property.value, self.max_length.value)

        if property.value is not None:
            value = f"{property.value}{value_ws}"
        else:
            value = value_ws

        comment = (
            f" # {property.comment}"
            if not str_is_empty_or_none(property.comment)
            else ""
        )

        yield f"{indent}{key}{value}{comment}".rstrip()

    def _serialize_datablock(self, datablock: Optional[Datablock]) -> Lines:
        if datablock is None or self.max_length.datablock is None:
            return []

        indent = " " * self._config.total_datablock_indent
        return (self._serialize_row(row, indent) for row in datablock)

    def _serialize_row(self, row: DatablockRow, indent: str) -> str:
        elem_spacing = " " * self.config.datablock_spacing
        elems = (self._serialize_row_element(elem, i) for elem, i in zip(row, count()))

        return indent + elem_spacing.join(elems).rstrip()

    def _serialize_row_element(self, elem: str, index: int) -> str:
        max_length = self.max_length.datablock[index]  # type: ignore
        whitespace = _get_offset_whitespace(elem, max_length)
        return elem + whitespace


class Serializer:
    """Serializer serializes Document to its corresponding lines."""

    def __init__(self, config: INISerializerConfig):
        """Creates a new Serializer with the provided configuration.

        Args:
            config (SerializerConfig): The configuration of this Serializer.
        """
        self._config = config

    def serialize(self, document: Document) -> Lines:
        """Serialize the provided document into an iterable of lines.

        Args:
            document (Document): The Document to serialize.

        Returns:
            Lines: An iterable returning each line of the serialized Document.
        """
        header_iterable = self._serialize_document_header(document.header_comment)

        serialize_section = lambda s: SectionSerializer.serialize(s, self._config)
        sections = (serialize_section(section) for section in document.sections)
        sections_with_spacing = Serializer._interweave(sections, [""])
        sections_iterable = chain.from_iterable(sections_with_spacing)

        return chain(header_iterable, sections_iterable)

    def _serialize_document_header(self, header: Iterable[CommentBlock]) -> Lines:
        delimiter = self._config.comment_delimiter
        serialize = lambda cb: _serialize_comment_block(cb, delimiter)
        blocks = (serialize(block) for block in header)
        blocks_with_spacing = Serializer._interweave(blocks, [""])

        return chain.from_iterable(blocks_with_spacing)

    @staticmethod
    def _interweave(iterable: Iterable, val: Any) -> Iterable:
        # Interweave the provided iterable with the provided value:
        # iterable_element, val, iterable_element, val, ...

        # Note that this will interweave with val without making copies
        # as such it is the same object being interweaved.
        return chain.from_iterable(zip(iterable, repeat(val)))


def write_ini(path: Path, document: Document, config: INISerializerConfig) -> None:
    """Write the provided document to the specified path

    If the provided path already exists, it will be overwritten. If the parent folder
    do not exist, they will be created.

    Args:
        path (Path): The path to which the document should be written.
        document (Document): The document to serialize to the specified path.
        config (INISerializerConfig): The configuration settings for the serializer.
    """

    serializer = Serializer(config)

    path.parent.mkdir(parents=True, exist_ok=True)

    with path.open("w", encoding="utf8") as f:

        for line in serializer.serialize(document):
            f.write(line + "\n")



from pathlib import Path
from typing import Any, Dict, List

from hydrolib.core.basemodel import (
    DiskOnlyFileModel,
    FileModel,
    ModelSaveSettings,
    SerializerConfig,
)
from hydrolib.core.dflowfm.extold.common_io import ORDERED_FORCING_FIELDS
from hydrolib.core.utils import FilePathStyleConverter


class Serializer:
    """Serializer class for serializing the forcing data of the `ExtOldModel` to file."""

    _file_path_style_converter = FilePathStyleConverter()

    @staticmethod
    def serialize(
        path: Path,
        data: Dict[str, List[Any]],
        config: SerializerConfig,
        save_settings: ModelSaveSettings,
    ) -> None:

        """
        Serialize the given data and write it to a file at the given path.

        This function may create a new file at the given path, or overwrite an existing file.

        Args:
            path (Path): The path to write the serialized data to.
            data (Dict[str, List[Any]]): The data to be serialized. The data should contain two keys:
                - 'comment' (List[str]): a list of the comments
                - 'forcing' (List[Dict[str, Any]]): a list of the external forcing data
            config (SerializerConfig): Configuration settings for the serializer.
            save_settings (ModelSaveSettings): Settings for how the model should be saved.
        """

        path.parent.mkdir(parents=True, exist_ok=True)

        serialized_comments: List[str] = []
        serialized_blocks: List[str] = []

        for comment in data["comment"]:
            serialized_comment = Serializer._serialize_comment(comment)
            serialized_comments.append(serialized_comment)

        for forcing in data["forcing"]:
            serialized_block = Serializer._serialize_forcing(
                forcing, config, save_settings
            )
            serialized_blocks.append(serialized_block)

        file_content: str = (
            "\n".join(serialized_comments) + "\n" + "\n\n".join(serialized_blocks)
        )

        with path.open("w", encoding="utf8") as f:
            f.write(file_content)

    @staticmethod
    def _serialize_comment(comment: str):
        return f"*{comment}"

    @staticmethod
    def _serialize_forcing(
        forcing: Dict[str, Any],
        config: SerializerConfig,
        save_settings: ModelSaveSettings,
    ) -> str:

        serialized_rows = []

        for key in ORDERED_FORCING_FIELDS:
            value = forcing.get(key.lower())

            if Serializer._skip_field_serialization(value):
                continue

            value = Serializer._convert_value(value, config, save_settings)

            serialized_row = f"{key}={value}"
            serialized_rows.append(serialized_row)

        serialized_block = "\n".join(serialized_rows)
        return serialized_block

    @classmethod
    def _convert_value(
        cls, value: Any, config: SerializerConfig, save_settings: ModelSaveSettings
    ) -> str:
        if isinstance(value, float):
            return f"{value:{config.float_format}}"
        if isinstance(value, FileModel):
            return Serializer._file_path_style_converter.convert_from_os_style(
                value.filepath, save_settings.path_style
            )

        return str(value)

    @classmethod
    def _skip_field_serialization(cls, value: Any) -> str:
        return value is None or (
            isinstance(value, DiskOnlyFileModel) and value.filepath is None
        )



from pathlib import Path
from typing import Iterable
from warnings import warn

from hydrolib.core.utils import get_substring_between


class NetworkTopologyFileParser:
    """A parser for RR topology files such as node and link files."""

    def __init__(self, enclosing_tag: str):
        """Initializes a new instance of the `NetworkTopologyFileParser` class.

        Args:
            enclosing_tag (str): The enclosing tag for the enclosed topology data per record.
        """

        self._enclosing_tag = enclosing_tag

    def parse(self, path: Path) -> dict:
        """Parses a network topology file to a dictionary.

        Args:
            path (Path): Path to the network topology file.
        """

        file_content = self._read_file(path)
        return self._parse_lines(file_content)

    def _read_file(self, path: Path) -> Iterable[str]:
        if not path.is_file():
            warn(f"File: `{path}` not found, skipped parsing.")
            return []

        with open(path, encoding="utf8") as file:
            lines = file.readlines()
        return lines

    def _parse_lines(self, lines: Iterable[str]) -> dict:
        records = []

        key_start = self._enclosing_tag.upper()
        key_end = self._enclosing_tag.lower()

        for line in lines:

            substring = get_substring_between(line, key_start, key_end)
            if substring == None:
                continue

            record = self._parse_line(substring)
            records.append(record)

        return {key_end: records}

    def _parse_line(self, line: str) -> dict:
        parts = line.split()

        record = {}

        index = 0
        while index < len(parts) - 1:
            key = parts[index]
            if key == "mt" and parts[index + 1] == "1":
                # `mt 1` is one keyword, but was parsed as two separate parts.
                index += 1

            index += 1
            value = parts[index].strip("'")
            index += 1

            record[key] = value

        return record



from pathlib import Path
from typing import Dict, List

from hydrolib.core.dflowfm.extold.common_io import ORDERED_FORCING_FIELDS


class Parser:
    """Parser class for parsing the forcing data of the old external forcings file to a dictionary to construct the `ExtOldModel` with."""

    @staticmethod
    def parse(filepath: Path) -> Dict:
        """Parses the file at the specified path to the forcing data.

        If a line starts with an asterisk (*) it is considered a comment.
        Comments are only allowed at the header of the file. Elsewhere, they will be skipped.

        Args:
            path (Path): The path to parse the data from.

        Returns:
            Dict[str, List[Any]]: A dictionary with the parsed data, with two keys:
                - 'comment' (List[str]): A list of the parsed comments.
                - 'forcing' (List[Dict[str, str]]): A list of the parsed forcing data as dictionaries. Each dictionary represents a forcing block from the file, with the parsed key value pairs.

        Raises:
            ValueError: Thrown when the order of a forcing block is not correct. Fields should be in the following order:
            "QUANTITY", "FILENAME", "VARNAME", "SOURCEMASK", "FILETYPE", "METHOD", "OPERAND", "VALUE", "FACTOR", "IFRCTYP",
            "AVERAGINGTYPE", "RELATIVESEARCHCELLSIZE", "EXTRAPOLTOL", "PERCENTILEMINMAX", "AREA", "NUMMIN"
        """

        with filepath.open(encoding="utf8") as file:
            lines = file.readlines()

        comments, start_data_index = Parser._parse_header(lines)
        forcings = Parser._parse_data(lines, start_data_index)

        return dict(comment=comments, forcing=forcings)

    @staticmethod
    def _parse_header(lines: List[str]):
        comments: List[str] = []

        start_data_index = 0
        for line_index in range(len(lines)):

            line = lines[line_index].strip()

            if len(line) == 0:
                comments.append(line)
                continue

            if line.startswith("*"):
                comments.append(line[1:])
                continue

            start_data_index = line_index
            break

        return comments, start_data_index

    @staticmethod
    def _parse_data(lines: List[str], start_index: int):
        forcings: List[Dict[str, str]] = []
        current_forcing: Dict[str, str] = {}

        for line_index in range(start_index, len(lines)):

            line = lines[line_index].strip()

            if line.startswith("*"):
                continue

            if len(line) == 0:
                if len(current_forcing) != 0:
                    Parser._validate_order(current_forcing, line_index)
                    forcings.append(current_forcing)
                    current_forcing = {}
                continue

            key, value = line.split("=", 1)
            current_forcing[key.strip()] = value.strip()

        if len(current_forcing) != 0:
            Parser._validate_order(current_forcing, line_index)
            forcings.append(current_forcing)

        return forcings

    @staticmethod
    def _validate_order(forcing: Dict[str, str], line_number: int):
        """Validates the order of the forcing fields given in the forcing block.

        - The fields are compared case insensitive.
        - For each KNOWN field that was parsed the order is checked.
        """

        parsed_fields_upper = [f.upper() for f in forcing.keys()]
        model_fields_upper = [f.upper() for f in ORDERED_FORCING_FIELDS]

        # Get the ordered KNOWN parsed fields, by filtering out unknown fields
        parsed_fields_ordered = [
            f for f in model_fields_upper if f in parsed_fields_upper
        ]

        # Get the unorderd KNOWN parsed fields, by filtering out unknown fields
        parsed_fields_unordered = [
            f for f in parsed_fields_upper if f in model_fields_upper
        ]

        if parsed_fields_unordered != parsed_fields_ordered:
            line_number_start = line_number + 1 - len(parsed_fields_upper)
            parsed_fields_ordered_str = ", ".join(parsed_fields_ordered)
            raise ValueError(
                f"Line {line_number_start}: Properties should be in the following order: {parsed_fields_ordered_str}"
            )



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from hydrolib.core.basemodel import ModelSaveSettings, SerializerConfig

TimeSeriesRow = List[str]
TimeSeriesBlock = List[TimeSeriesRow]


class TimSerializerConfig(SerializerConfig):
    """Configuration settings for the TimSerializer."""

    column_spacing: int = 1
    """(int): The number of spaces to include between columns in the serialized .tim file."""


class TimSerializer:
    @staticmethod
    def serialize(
        path: Path,
        data: Dict[str, List[Any]],
        config: TimSerializerConfig,
        save_settings: ModelSaveSettings,
    ) -> None:
        """
        Serialize timeseries data to a file in .tim format.

        Args:
            path (Path): The path to the destination .tim file.
            data (Dict[str, List[Any]]): A dictionary with keys "comments" and "timeseries".
            - "comments" represents comments found at the start of the file.
            - "timeseries" is a list of dictionaries with the key as "time" and values as "data".
                - "time" is a time as a string.
                - "data" is data as a list of strings.

            config (TimSerializerConfig): The serialization configuration settings.
            save_settings (ModelSaveSettings): The save settings to be used.
        """
        path.parent.mkdir(parents=True, exist_ok=True)

        commentlines = TimSerializer._serialize_comment_lines(data)
        timeserieslines = TimSerializer._serialize_timeseries_lines(data, config)

        file_content = TimSerializer._serialize_file_content(
            timeserieslines, commentlines
        )
        with path.open("w", encoding="utf8") as file:
            file.write(file_content)

    @staticmethod
    def _serialize_comment_lines(data: Dict[str, List[Any]]) -> List[str]:
        commentlines = []
        for comment in data["comments"]:
            commentlines.append(f"#{comment}")
        return commentlines

    @staticmethod
    def _serialize_timeseries_lines(
        data: Dict[str, List[Any]], config: TimSerializerConfig
    ) -> List[str]:
        format_float = lambda v: f"{v:{config.float_format}}"
        timeseriesblock = TimSerializer._serialize_to_timeseries_block(
            data, format_float
        )
        timeserieslines = TimSerializer._serialize_timeseries_to_lines(
            timeseriesblock, config
        )
        return timeserieslines

    @staticmethod
    def _serialize_to_timeseries_block(
        data: Dict[str, List[Any]], format_float: Callable[[float], str]
    ) -> TimeSeriesBlock:
        timeseries_block: TimeSeriesBlock = []
        for timeseries in data["timeseries"]:
            time = timeseries["time"]
            row_elements = timeseries["data"]
            timeseries_row = [format_float(time)] + [
                format_float(value) for value in row_elements
            ]
            timeseries_block.append(timeseries_row)
        return timeseries_block

    @staticmethod
    def _serialize_timeseries_to_lines(
        timeseries_block: TimeSeriesBlock, config: TimSerializerConfig
    ) -> List[str]:
        # Make sure the columns are aligned and have the proper spacing
        column_space = " " * config.column_spacing
        column_lengths = TimSerializer._get_column_lengths(timeseries_block)

        timeserieslines = []
        for timeseries_row in timeseries_block:
            row_elements: List[str] = []
            for index, value in enumerate(timeseries_row):
                whitespace_offset = TimSerializer._get_offset_whitespace(
                    value, column_lengths[index]
                )
                row_elements.append(value + whitespace_offset)

            line = column_space.join(row_elements)
            timeserieslines.append(line)
        return timeserieslines

    @staticmethod
    def _get_offset_whitespace(value: Optional[str], max_length: int) -> str:
        value_length = len(value) if value is not None else 0
        return " " * max(max_length - value_length, 0)

    @staticmethod
    def _serialize_file_content(timeserieslines: List[str], commentlines: List[str]):
        lines = []
        lines.extend(commentlines)
        lines.extend(timeserieslines)
        file_content = "\n".join(lines)
        return file_content

    @staticmethod
    def _get_column_lengths(timeseries_block: TimeSeriesBlock) -> List[int]:
        if len(timeseries_block) == 0:
            return []

        n_columns = len(timeseries_block[0])
        column_lengths = [0] * n_columns

        for timeseries_row in timeseries_block:
            for index, row_element in enumerate(timeseries_row):
                if len(row_element) > column_lengths[index]:
                    column_lengths[index] = len(row_element)

        return column_lengths



from pathlib import Path
from warnings import warn

from lxml import etree


class DIMRParser:
    """A parser for DIMR xml files."""

    @staticmethod
    def parse(path: Path) -> dict:
        """Parses a DIMR file to a dictionary.

        Args:
            path (Path): Path to the DIMR configuration file.
        """
        if not path.is_file():
            warn(f"File: `{path}` not found, skipped parsing.")
            return {}

        parser = etree.XMLParser(
            remove_comments=True, resolve_entities=False, no_network=True
        )
        root = etree.parse(str(path), parser=parser).getroot()

        return DIMRParser._node_to_dictionary(root, True)

    @staticmethod
    def _node_to_dictionary(node: etree, ignore_attributes: bool = False):
        """
        Convert an lxml.etree node tree recursively into a nested dictionary.
        The node's attributes and child items will be added to it's dictionary.

        Args:
            node (etree): The etree node
            ignore_attributes (bool): Optional parameter; whether or not to
                                      skip the node's attributes. Default is False.
        """

        result = {} if ignore_attributes else dict(node.attrib)

        for child_node in node.iterchildren():

            key = child_node.tag.split("}")[1]

            if child_node.text and child_node.text.strip():
                value = child_node.text
            else:
                value = DIMRParser._node_to_dictionary(child_node)

            if key in result:

                if type(result[key]) is list:
                    result[key].append(value)
                else:
                    first_value = result[key].copy()
                    result[key] = [first_value, value]
            else:
                result[key] = value

        return result



from .models import (
    Bridge,
    Compound,
    Culvert,
    CulvertSubType,
    Dambreak,
    DambreakAlgorithm,
    FlowDirection,
    GateOpeningHorizontalDirection,
    GeneralStructure,
    Orientation,
    Orifice,
    Pump,
    Structure,
    StructureGeneral,
    StructureModel,
    UniversalWeir,
    Weir,
)

__all__ = [
    "Structure",
    "FlowDirection",
    "Orientation",
    "Weir",
    "UniversalWeir",
    "CulvertSubType",
    "Culvert",
    "Pump",
    "Compound",
    "Orifice",
    "GateOpeningHorizontalDirection",
    "GeneralStructure",
    "DambreakAlgorithm",
    "Dambreak",
    "Bridge",
    "StructureGeneral",
    "StructureModel",
]



import os
from pathlib import Path
from typing import Any, Dict

from hydrolib.core.basemodel import ModelSaveSettings, SerializerConfig


class NodeFileSerializer:
    """Serializer for the RR node topology data."""

    @staticmethod
    def serialize(
        path: Path,
        data: dict,
        config: SerializerConfig,
        save_settings: ModelSaveSettings,
    ):
        """
        Serializes the RR node topology data to the file at the specified path.

        Args:
            path (Path): The path to the destination file.
            data (Dict): The data to be serialized.
            config (SerializerConfig): The serialization configuration.
            save_settings (ModelSaveSettings): The model save settings.
        """

        path.parent.mkdir(parents=True, exist_ok=True)

        with path.open("w", encoding="utf8") as f:
            for node in data["node"]:
                line = f"NODE {NodeFileSerializer._to_line(node, config)} node\n"
                f.write(line)

    @staticmethod
    def _to_line(node: Dict[str, Any], config: SerializerConfig) -> str:

        identifier = node["id"]
        nm = node["nm"]
        ri = node["ri"]
        mt = node["mt"]
        nt = node["nt"]
        obid = node["ObID"]
        px = node["px"]
        py = node["py"]

        float_format = lambda v: f"{v:{config.float_format}}"
        return f"id '{identifier}' nm '{nm}' ri '{ri}' mt 1 '{mt}' nt {nt} ObID '{obid}' px {float_format(px)} py {float_format(py)}"


class LinkFileSerializer:
    """Serializer for the RR link topology data."""

    @staticmethod
    def serialize(
        path: Path,
        data: dict,
        config: SerializerConfig,
        save_settings: ModelSaveSettings,
    ):
        """
        Serializes the RR link topology data to the file at the specified path.

        Args:
            path (Path): The path to the destination file.
            data (Dict): The data to be serialized.
            config (SerializerConfig): The serialization configuration.
            save_settings (ModelSaveSettings): The model save settings.
        """

        path.parent.mkdir(parents=True, exist_ok=True)

        with path.open("w", encoding="utf8") as f:
            for link in data["link"]:
                line = f"BRCH {LinkFileSerializer._to_line(link)} brch\n"
                f.write(line)

    @staticmethod
    def _to_line(link: Dict[str, Any]) -> str:
        identifier = link["id"]
        nm = link["nm"]
        ri = link["ri"]
        mt = link["mt"]
        bt = link["bt"]
        obid = link["ObID"]
        bn = link["bn"]
        en = link["en"]

        return f"id '{identifier}' nm '{nm}' ri '{ri}' mt 1 '{mt}' bt {bt} ObID '{obid}' bn '{bn}' en '{en}'"



from pathlib import Path
from typing import Callable, Dict, List, Optional

from pydantic import validator

from hydrolib.core.basemodel import (
    BaseModel,
    ModelSaveSettings,
    ParsableFileModel,
    SerializerConfig,
)
from hydrolib.core.utils import str_is_empty_or_none

from .parser import XYNParser
from .serializer import XYNSerializer


class XYNPoint(BaseModel):
    """Single XYN point, representing a named station location."""

    x: float
    """float: The x or λ coordinate."""

    y: float
    """float: The y or φ coordinate."""

    n: str
    """float: The name of the point."""

    def _get_identifier(self, data: dict) -> Optional[str]:
        x = data.get("x")
        y = data.get("y")
        n = data.get("n")
        return f"x:{x} y:{y} n:{n}"

    @validator("n", pre=True)
    def _validate_name(cls, value):
        if str_is_empty_or_none(value):
            raise ValueError("Name cannot be empty.")

        if "'" in value or '"' in value:
            raise ValueError(
                "Name cannot contain single or double quotes except at the start and end."
            )

        return value


class XYNModel(ParsableFileModel):
    """Observation station (.xyn) file."""

    points: List[XYNPoint] = []
    """List[`XYNPoint`]: List of XYN points."""

    def dict(self, *args, **kwargs):
        # speed up serializing by not converting these lowest models to dict
        return dict(points=self.points)

    @classmethod
    def _filename(cls) -> str:
        return "stations_obs"

    @classmethod
    def _ext(cls) -> str:
        return ".xyn"

    @classmethod
    def _get_serializer(
        cls,
    ) -> Callable[[Path, Dict, SerializerConfig, ModelSaveSettings], None]:
        return XYNSerializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable[[Path], Dict]:
        return XYNParser.parse



"""models.py defines all classes and functions related to representing pol/pli(z) files.
"""

from typing import Callable, List, Optional, Sequence

from hydrolib.core.basemodel import BaseModel, ModelSaveSettings, ParsableFileModel


class Description(BaseModel):
    """Description of a single PolyObject.

    The Description will be prepended to a block. Each line will
    start with a '*'.

    Attributes:
        content (str): The content of this Description.
    """

    content: str


class Metadata(BaseModel):
    """Metadata of a single PolyObject.

    Attributes:
        name (str): The name of the PolyObject
        n_rows (int): The number of rows (i.e. Point instances) of the PolyObject
        n_columns (int): The total number of values in a Point, including x, y, and z.
    """

    name: str
    n_rows: int
    n_columns: int


class Point(BaseModel):
    """Point consisting of a x and y coordinate, an optional z coordinate and data.

    Attributes:
        x (float): The x-coordinate of this Point
        y (float): The y-coordinate of this Point
        z (Optional[float]): An optional z-coordinate of this Point.
        data (Sequence[float]): The additional data variables of this Point.
    """

    x: float
    y: float
    z: Optional[float]
    data: Sequence[float]

    def _get_identifier(self, data: dict) -> Optional[str]:
        x = data.get("x")
        y = data.get("y")
        z = data.get("z")
        return f"x:{x} y:{y} z:{z}"


class PolyObject(BaseModel):
    """PolyObject describing a single block in a poly file.

    The metadata should be consistent with the points:
    - The number of points should be equal to number of rows defined in the metadata
    - The data of each point should be equal to the number of columns defined in the
      metadata.

    Attributes:
        description (Optional[Description]):
            An optional description of this PolyObject
        metadata (Metadata):
            The Metadata of this PolObject, describing the structure
        points (List[Point]):
            The points describing this PolyObject, structured according to the Metadata
    """

    description: Optional[Description]
    metadata: Metadata
    points: List[Point]


class PolyFile(ParsableFileModel):
    """Poly-file (.pol/.pli/.pliz) representation."""

    has_z_values: bool = False
    objects: Sequence[PolyObject] = []

    def _serialize(self, _: dict, save_settings: ModelSaveSettings) -> None:
        from .serializer import write_polyfile

        # We skip the passed dict for a better one.
        write_polyfile(self._resolved_filepath, self.objects, self.serializer_config)

    @classmethod
    def _ext(cls) -> str:
        return ".pli"

    @classmethod
    def _filename(cls) -> str:
        return "objects"

    @classmethod
    def _get_serializer(cls) -> Callable:
        # Unused, but requires abstract implementation
        pass

    @classmethod
    def _get_parser(cls) -> Callable:
        # TODO Prevent circular dependency in Parser
        from .parser import read_polyfile

        return read_polyfile



import platform
import re
from enum import Enum, auto
from operator import eq, ge, gt, le, lt, ne
from pathlib import Path
from typing import Any, Callable, List, Optional


def example(a: float, b: float = 1.0) -> float:
    """[summary]

    Args:
        a (float): [description]
        b (float): [description]

    Returns:
        float: [description]
    """
    return a * b


def to_key(string: str) -> str:
    """
    Construct a key name from a given field name.
    The given field name may be a Pydantic Field alias, and the key name
    is intended to be used as a BaseModel class member variable name.

    Args:
        string (str): input field name

    """
    # First replace any leading digits, because those are undesirable
    # in variable names.
    digitdict = {
        "0": "zero",
        "1": "one",
        "2": "two",
        "3": "three",
        "4": "four",
        "5": "five",
        "6": "six",
        "7": "seven",
        "8": "eight",
        "9": "nine",
    }
    m = re.search(r"^\d+", string)
    if m:
        digitstring = string[0 : m.end()]
        for key, val in digitdict.items():
            digitstring = digitstring.replace(key, val)
        string = digitstring + string[m.end() :]

    # Next, replace spaces and hyphens in the potential variable name.
    return string.lower().replace(" ", "_").replace("-", "")


def to_list(item: Any) -> List[Any]:
    """Puts the specified item in a list if it is an instance of `dict`.

    Attributes:
        item: The item to put in a list.

    Returns:
        List: A list with the specified item.
    """

    if not isinstance(item, list):
        return [item]
    return item


def str_is_empty_or_none(str_field: str) -> bool:
    """
    Verifies whether a string is empty or None.

    Args:
        str_field (str): String to validate.

    Returns:
        bool: Evaluation result.
    """
    return str_field is None or not str_field or str_field.isspace()


def get_str_len(str_field: Optional[str]) -> int:
    """
    Get string length or 0 if input is None.

    Args:
        str_field (str): String to measure.

    Returns:
        int: Length of passed input.
    """
    return len(str_field) if str_field else 0


def get_substring_between(source: str, start: str, end: str) -> Optional[str]:
    """Finds the substring between two other strings.

    Args:
        source (str): The source string.
        start (str): The starting string from where to create the substring.
        end (str): The end string to where to create the substring.

    Returns:
        str: The substring if found; otherwise, `None`.
    """

    index_start = source.find(start)
    if index_start == -1:
        return None

    index_start += len(start)

    index_end = source.rfind(end, index_start)
    if index_end == -1:
        return None

    return source[index_start:index_end]


def operator_str(operator_func: Callable) -> str:
    """
    Make string representation of some of operator's built-in operator
    functions, for use in prettyprinting.

    Args:
        operator_func (Callable): Typically one of operator's built-in
            operator functions. When unsupported, the standard __str__
            representation is returned.
    """
    if operator_func == eq:
        return "is"
    elif operator_func == ne:
        return "is not"
    elif operator_func == lt:
        return "is less than"
    elif operator_func == le:
        return "is less than or equal to"
    elif operator_func == gt:
        return "is greater than"
    elif operator_func == ge:
        return "is greater than or equal to"
    else:
        return str(operator_func)


class OperatingSystem(Enum):
    """Enum represting an operating system."""

    WINDOWS = auto()
    """The Windows operating system."""
    LINUX = auto()
    """The Linux operating system."""
    MACOS = auto()
    """The MacOS operating system."""


def get_operating_system() -> OperatingSystem:
    """Gets the currently running operating system.

    Raises:
        NotImplementedError: When the current operating system is anything other than Windows, Linux or MacOS.

    Returns:
        OperatingSystem: The operating system.
    """
    operating_system = platform.system()

    if operating_system == "Windows":
        return OperatingSystem.WINDOWS
    if operating_system == "Linux":
        return OperatingSystem.LINUX
    if operating_system == "Darwin":
        return OperatingSystem.MACOS

    raise NotImplementedError(f"Operating system {operating_system} is not supported.")


class PathStyle(str, Enum):
    """Path style format."""

    UNIXLIKE = "unix"
    """Unix-like path style."""
    WINDOWSLIKE = "windows"
    """Windows-like path style."""


def get_path_style_for_current_operating_system() -> PathStyle:
    """Gets the file path style for the currently running operating system.

    Returns:
        OperatingSystem: The OS path style.
    """
    operating_system = get_operating_system()

    if operating_system == OperatingSystem.WINDOWS:
        return PathStyle.WINDOWSLIKE
    else:
        return PathStyle.UNIXLIKE


class FilePathStyleConverter:
    """Class for converting file paths between different path styles."""

    _os_path_style = get_path_style_for_current_operating_system()

    def convert_to_os_style(self, file_path: Path, source_path_style: PathStyle) -> str:
        """Convert the file path from the source path style to the path style of the current operating system.

        Args:
            file_path (Path): The file path to convert to the OS path style.
            file_path_style (PathStyle): The file path style of the given file path.

        Returns:
            str: The converted file path with OS path style.

        Raises:
            NotImplementedError: When this function is called with a PathStyle other than WINDOWSLIKE or UNIXLIKE.
        """

        return FilePathStyleConverter._convert(
            file_path, source_path_style, self._os_path_style
        )

    def convert_from_os_style(
        self, file_path: Path, target_path_style: PathStyle
    ) -> str:
        """Convert the file path from the path style of the current operating system to the target path style.

        Args:
            file_path (Path): The file path to convert to the target path style.
            target_path_style (PathStyle): The target path style to convert the file path to.

        Returns:
            str: The converted file path with the target path style.

        Raises:
            NotImplementedError: When this function is called with a PathStyle other than WINDOWSLIKE or UNIXLIKE.
        """

        return FilePathStyleConverter._convert(
            file_path, self._os_path_style, target_path_style
        )

    @classmethod
    def _convert(
        cls, file_path: Path, source_path_style: PathStyle, target_path_style: PathStyle
    ) -> str:
        if source_path_style == target_path_style:
            return file_path.as_posix()

        if (
            source_path_style == PathStyle.UNIXLIKE
            and target_path_style == PathStyle.WINDOWSLIKE
        ):
            return FilePathStyleConverter._from_posix_to_windows_path(file_path)
        elif (
            source_path_style == PathStyle.WINDOWSLIKE
            and target_path_style == PathStyle.UNIXLIKE
        ):
            return FilePathStyleConverter._from_windows_to_posix_path(file_path)
        else:
            raise NotImplementedError(
                f"Cannot convert {source_path_style} to {target_path_style}"
            )

    @classmethod
    def _from_posix_to_windows_path(cls, posix_path: Path) -> str:
        is_relative = not posix_path.as_posix().startswith("/")

        if is_relative:
            return posix_path.as_posix()

        root = posix_path.parts[1]
        windows_root = root + ":/"
        parts = posix_path.parts[2:]

        windows_path = windows_root + "/".join(parts)

        return windows_path

    @classmethod
    def _from_windows_to_posix_path(cls, windows_path: Path) -> str:
        windows_path_str = str(windows_path).replace("\\", "/")
        windows_path = Path(windows_path_str)

        root = windows_path.parts[0]
        is_relative = ":" not in root

        if is_relative:
            return windows_path_str

        posix_root = "/" + root.split(":")[0] + "/"
        parts = windows_path.parts[1:]

        posix_path = posix_root + "/".join(parts)

        return posix_path



import re
from pathlib import Path
from typing import Dict

xyzpattern = re.compile(r"\s+")


class XYZParser:
    """
    A parser for .xyz files which are like this:

    number number    number
    number number number # comment

    Note that the whitespace can vary and the comment
    left out.
    """

    @staticmethod
    def parse(filepath: Path) -> Dict:
        """Parse an .xyz file into a Dict with the list of points read.

        Args:
            filepath (Path): .xyz file to be read.

        Returns:
            Dict: dictionary with "points" value set to a list of points
                each of which is a dict itself, with keys 'x', 'y', 'z'
                and 'c' for an optional comment.

        Raises:
            ValueError: if a line in the file contains no values that
                could be parsed.
        """

        data: Dict = dict(points=[])

        with filepath.open(encoding="utf8") as f:
            for linenr, line in enumerate(f.readlines()):

                line = line.strip()
                if line.startswith("*") or len(line) == 0:
                    continue

                try:
                    x, y, z, *c = re.split(xyzpattern, line, maxsplit=3)
                except ValueError:
                    raise ValueError(
                        f"Error parsing XYZ file '{filepath}', line {linenr+1}."
                    )

                c = c[0] if len(c) > 0 else ""
                c = c.strip("#").strip()
                if len(c) == 0:
                    c = None

                data["points"].append(dict(x=x, y=y, z=z, comment=c))

        return data



from pathlib import Path
from typing import Any, Dict, List, Tuple

TimData = Dict[str, List[str]]


class TimParser:
    """
    A parser for .tim files.
    Full line comments at the start of the file are supported. Comment lines start with either a `*` or a `#`.
    No other comments are supported.
    """

    @staticmethod
    def parse(filepath: Path) -> Dict[str, List[Any]]:
        """Parse a .tim file into a dictionary with comments and time series data.

        Args:
            filepath (Path): Path to the .tim file to be parsed.

        Returns:
            Dict[str, List[Any]]: A dictionary with keys "comments" and "timeseries".
            - "comments" represents comments found at the start of the file.
            - "timeseries" is a list of dictionaries with the key as "time" and values as "data".
                - "time" is a time as a string.
                - "data" is data as a list of strings.

        Raises:
            ValueError: If the file contains a comment that is not at the start of the file.
            ValueError: If the data of the timeseries is empty.
        """

        comments: List[str] = []
        timeseries: List[TimData] = []

        with filepath.open(encoding="utf8") as file:
            lines = file.readlines()
            comments, start_timeseries_index = TimParser._read_header_comments(lines)
            timeseries = TimParser._read_time_series_data(lines, start_timeseries_index)

        return {"comments": comments, "timeseries": timeseries}

    @staticmethod
    def _read_header_comments(lines: List[str]) -> Tuple[List[str], int]:
        """Read the header comments of the lines from the .tim file.
        The comments are only expected at the start of the .tim file.
        When a non comment line is encountered, all comments from the header will be retuned together with the start index of the timeseries data.

        Args:
            lines (List[str]): Lines from the the .tim file which is read.

        Returns:
            Tuple of List[str] and int, the List[str] contains the commenst from the header, the int is the start index of the timeseries.
        """
        comments: List[str] = []
        start_timeseries_index = 0
        for line_index in range(len(lines)):

            line = lines[line_index].strip()

            if len(line) == 0:
                comments.append(line)
                continue

            if line.startswith("#") or line.startswith("*"):
                comments.append(line[1:])
                continue

            start_timeseries_index = line_index
            break

        return comments, start_timeseries_index

    @staticmethod
    def _read_time_series_data(
        lines: List[str], start_timeseries_index: int
    ) -> List[TimData]:
        timeseries: List[TimData] = []
        for line_index in range(start_timeseries_index, len(lines)):
            line = lines[line_index].strip()

            if len(line) == 0:
                continue

            TimParser._raise_error_if_contains_comment(line, line_index + 1)

            time, *values = line.split()

            TimParser._raise_error_if_values_empty(values, line_index)

            timrecord = {"time": time, "data": values}
            timeseries.append(timrecord)

        return timeseries

    @staticmethod
    def _raise_error_if_contains_comment(line: str, line_index: int) -> None:
        if "#" in line or "*" in line:
            raise ValueError(
                f"Line {line_index}: comments are only supported at the start of the file, before the time series data."
            )

    @staticmethod
    def _raise_error_if_values_empty(values: List[str], line_index: int) -> None:
        if len(values) == 0:
            raise ValueError(f"Line {line_index}: Time series cannot be empty.")



from pathlib import Path
from typing import Dict

from hydrolib.core.basemodel import ModelSaveSettings, SerializerConfig


class XYNSerializer:
    @staticmethod
    def serialize(
        path: Path,
        data: Dict,
        config: SerializerConfig,
        save_settings: ModelSaveSettings,
    ) -> None:
        """
        Serializes the observation point data to an .xyn file at the specified path.

        If the name contains spaces, it will be surrounded with single quotes.

        Args:
            path (Path): The path to the destination file.
            data (Dict[str, List[XYNPoint]]): The data to be serialized.
                The dictionary should contain a single key 'points' that holds a list of XYNPoints.
            config (SerializerConfig): The serialization configuration.
            save_settings (ModelSaveSettings): The model save settings.
        """
        path.parent.mkdir(parents=True, exist_ok=True)

        space = 1 * " "
        format_float = lambda x: f"{x:{config.float_format}}"
        format_name = lambda n: f"'{n}'" if " " in n else n

        serialized_points = []

        for point in data["points"]:
            serialized_point: str = space.join(
                [
                    format_float(point.x),
                    format_float(point.y),
                    format_name(point.n),
                ]
            )
            serialized_points.append(serialized_point)

        file_content: str = "\n".join(serialized_points)

        with path.open("w", encoding="utf8") as f:
            f.write(file_content)



from enum import Enum
from typing import Dict, List, Literal, Optional

from pydantic.class_validators import root_validator, validator
from pydantic.fields import Field

from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    get_enum_validator,
    get_split_string_on_delimiter_validator,
    make_list_validator,
    validate_correct_length,
    validate_required_fields,
)


class NodeType(str, Enum):
    """
    Enum class containing the valid values for the node type
    as used in StorageNode.
    """

    inspection = "inspection"
    soakawaydrain = "soakawayDrain"
    compartment = "compartment"
    unspecified = "unspecified"


class StorageType(str, Enum):
    """
    Enum class containing the valid values for the storage type
    as used in StorageNode.
    """

    reservoir = "reservoir"
    closed = "closed"


class Interpolation(str, Enum):
    """
    Enum class containing the valid values for the interpolation type
    as used for a storage area table in StorageNode.
    """

    linear = "linear"
    block = "block"


class StorageNodeGeneral(INIGeneral):
    """The storage node file's `[General]` section with file meta data."""

    class Comments(INIBasedModel.Comments):
        fileversion: Optional[str] = Field(
            "File version. Do not edit this.", alias="fileVersion"
        )
        filetype: Optional[str] = Field(
            "File type. Should be 'storageNodes'. Do not edit this.",
            alias="fileType",
        )
        usestreetstorage: Optional[str] = Field(
            "Indicates whether above streetLevel the streetStorageArea must be used (true) "
            + "or the regular storage area continues (false). This option is only applicable "
            + "for storage nodes with useTable = false.",
            alias="useStreetStorage",
        )

    comments: Comments = Comments()
    _header: Literal["General"] = "General"
    fileversion: str = Field("2.00", alias="fileVersion")
    filetype: Literal["storageNodes"] = Field("storageNodes", alias="fileType")
    usestreetstorage: Optional[bool] = Field(True, alias="useStreetStorage")


class StorageNode(INIBasedModel):
    """
    A storage node that is included in the storage node file.

    All lowercased attributes match with the storage node input as described in
    [UM Sec.C.17](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#section.C.17).
    """

    class Comments(INIBasedModel.Comments):
        id: Optional[str] = Field("Unique id of the storage node.", alias="id")
        name: Optional[str] = Field("Long name in the user interface.", alias="name")
        manholeid: Optional[str] = Field(
            "(optional) Unique id of manhole that this (compartment) node is part of.",
            alias="manholeId",
        )
        nodetype: Optional[str] = Field(
            "(optional) Type of the node. Possible values are: "
            + "inspection: inspection chamber, "
            + "soakawayDrain: soakaway drain (infiltration), "
            + "compartment: manhole compartment, "
            + "unspecified: general storage node of unspecified type",
            alias="nodeType",
        )
        nodeid: Optional[str] = Field(
            "Connection node on which the storage node is located.", alias="nodeId"
        )
        usetable: Optional[str] = Field(
            "Switch to select a simple (false) or tabulated (true) storage area.",
            alias="useTable",
        )
        bedlevel: Optional[str] = Field(
            "Bed level of the storage area [m AD].", alias="bedLevel"
        )
        area: Optional[str] = Field(
            "Storage area from bedLevel up to streetLevel (and beyond if useStreetStorage = false) [m2].",
            alias="area",
        )
        streetlevel: Optional[str] = Field(
            "Street level of the storage area [m AD].", alias="streetLevel"
        )
        streetstoragearea: Optional[str] = Field(
            "Storage area from streetLevel upwards if useStreetStorage = true [m2].",
            alias="streetStorageArea",
        )
        storagetype: Optional[str] = Field(
            "Possible values: "
            + "reservoir: Above streetLevel the storage area of this node is also taken into account. "
            + "closed: Above streetLevel this storage node has no storage area.",
            alias="storageType",
        )
        numlevels: Optional[str] = Field(
            "Number of levels in storage area table.", alias="numLevels"
        )
        levels: Optional[str] = Field(
            "Levels in storage area table [m].", alias="levels"
        )
        storagearea: Optional[str] = Field(
            "Areas in storage area table [m2].", alias="storageArea"
        )
        interpolate: Optional[str] = Field(
            "Interpolation type for storage area table. Possible values: linear or block.",
            alias="interpolate",
        )

    comments: Comments = Comments()
    _header: Literal["StorageNode"] = "StorageNode"
    id: str = Field(alias="id")
    name: str = Field(alias="name")
    manholeid: Optional[str] = Field(alias="manholeId")

    nodetype: Optional[NodeType] = Field(NodeType.unspecified.value, alias="nodeType")
    nodeid: str = Field(alias="nodeId")
    usetable: bool = Field(alias="useTable")

    # useTable is True
    bedlevel: Optional[float] = Field(alias="bedLevel")
    area: Optional[float] = Field(alias="area")
    streetlevel: Optional[float] = Field(alias="streetLevel")
    streetstoragearea: Optional[float] = Field(alias="streetStorageArea")
    storagetype: Optional[StorageType] = Field(
        StorageType.reservoir.value, alias="storageType"
    )

    # useTable is False
    numlevels: Optional[int] = Field(alias="numLevels")
    levels: Optional[List[float]] = Field(alias="levels")
    storagearea: Optional[List[float]] = Field(alias="storageArea")
    interpolate: Optional[Interpolation] = Field(
        Interpolation.linear.value, alias="interpolate"
    )

    _interpolation_validator = get_enum_validator("interpolate", enum=Interpolation)
    _nodetype_validator = get_enum_validator("nodetype", enum=NodeType)
    _storagetype_validator = get_enum_validator("storagetype", enum=StorageType)
    _split_to_list = get_split_string_on_delimiter_validator(
        "levels",
        "storagearea",
    )

    @root_validator(allow_reuse=True)
    def check_list_length_levels(cls, values):
        """Validates that the length of the levels field is as expected."""
        return validate_correct_length(
            values,
            "levels",
            "storagearea",
            length_name="numlevels",
            list_required_with_length=True,
        )

    @root_validator(allow_reuse=True)
    def validate_that_required_fields_are_present_when_using_tables(
        cls, values: Dict
    ) -> Dict:
        """Validates that the specified fields are present when the usetable field is also present."""
        return validate_required_fields(
            values,
            "numlevels",
            "levels",
            "storagearea",
            conditional_field_name="usetable",
            conditional_value=True,
        )

    @root_validator(allow_reuse=True)
    def validate_that_required_fields_are_present_when_not_using_tables(
        cls, values: Dict
    ) -> Dict:
        """Validates that the specified fields are present."""
        return validate_required_fields(
            values,
            "bedlevel",
            "area",
            "streetlevel",
            conditional_field_name="usetable",
            conditional_value=False,
        )

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("id") or data.get("name")


class StorageNodeModel(INIModel):
    """
    The overall storage node model that contains the contents of one storage node file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.storagenodefile[..]`.

    Attributes:
        general (StorageNodeGeneral): `[General]` block with file metadata.
        storagenode (List[StorageNode]): List of `[StorageNode]` blocks for all storage nodes.
    """

    general: StorageNodeGeneral = StorageNodeGeneral()
    storagenode: List[StorageNode] = []

    _make_list = make_list_validator("storagenode")

    @classmethod
    def _filename(cls) -> str:
        return "nodeFile"

    @validator("storagenode", each_item=True)
    def _validate(cls, storagenode: StorageNode, values: dict):
        """Validates for each storage node whether the streetStorageArea value is provided
        when the general useStreetStorage is True and the storage node useTable is False.
        """

        usestreetstorage = values["general"].usestreetstorage

        if (
            usestreetstorage
            and not storagenode.usetable
            and storagenode.streetstoragearea is None
        ):
            raise ValueError(
                f"streetStorageArea should be provided when useStreetStorage is True and useTable is False for storage node with id {storagenode.id}"
            )

        return storagenode
